local __ = require '__'
local log = require 'log'

local ev = require 'ev'
local ffi = require 'ffi'
local ffi_cast = ffi.cast
local ffi_copy = ffi.copy
local ffi_new = ffi.new
local ffi_string = ffi.string
local ts = require 'timestamps'
local u = require 'unix'
local log = require 'log'
local config = require 'system_config'
local armada
if config.device_type == 'edaqxr' then
    armada = require 'armada'
end

local sample_clock = require 'sample_clock'

local channel_proxy = require 'channel_proxy'
local cca_handler = require 'cca_handler'
local message_can = require 'message_can'

local can = require 'can'

local C = ffi.C

require 'strict'

local update_rate = 10

local dumper = require 'data_dumper'

local can_buf_count = 1024

ffi.cdef[[
    ssize_t write(int fd, const void *buf, size_t count);

    typedef struct __attribute__((packed)) can_sink {
        uint64_t base_timestamp;
        int socket;
        int max_count;
        int count;
        struct {
            uint32_t len;
            message_can msg[1024];
        } buf;
        char    *b;
    } can_sink;

    char *strerror(int errnum);
    uint32_t htonl(uint32_t hostlog);
    int usleep(uint32_t usec);
]]

local out_of_space_msg = "Out of buffer space for CAN messages to be sent " ..
    "to the test engine.  This error indicates that the test engine is not " ..
    "able to keep up with reading the CAN message data.  A common cause " ..
    "for this is using a start time delay that is too short for the test.  " ..
    "Increase the start time delay for the test and try again.  This error " ..
    "causes the flagship proxy to exit and the system must be rebooted " ..
    "to restart it."

local htonll_data = ffi.new('union htonll_u')
if ffi.abi("be") then -- nothing to do
    function htonll(b) return b end
else
    function htonll(b)    -- assumes b is ffi C type long long
        htonll_data.li = b
        local i1 = C.htonl(htonll_data.i[1])
        local i0 = C.htonl(htonll_data.i[0])
        htonll_data.i[1] = i0
        htonll_data.i[0] = i1
        return htonll_data.li
    end
end

local can_sink = ffi.metatype('can_sink', { __index = {
    flush = function(self, pkt)
        if self.count <= 0 then
            return
        end
        local towrite
        self.buf.len = self.count * ffi.sizeof'message_can'
        towrite = ffi.sizeof'uint32_t' + self.buf.len
        self.buf.len = C.htonl(self.buf.len)
        self.b = ffi.cast('char *', self.buf)

        local max_attempts = 100
        while towrite > 0 and max_attempts > 0 do
            local wrote, estr, e =
                u.write(self.socket, ffi_string(self.b, towrite))
            if not wrote then
                if e == u.EAGAIN then
                    -- Don't do anything other than retry the write in 1ms
                    C.usleep(1000)
                    max_attempts = max_attempts - 1
                elseif e ~= u.EPIPE and e ~= u.ECONNRESET then
                    log:log(log.ERR,
                        string.format("Write error on CAN sink: %s", estr))
                    os.exit(-5)
                else
                    -- Connection is gone, ignore the error
                    return
                end
            else
                self.b = self.b + wrote
                towrite = towrite - wrote
            end
        end
        if max_attempts <= 0 then
            log:log(log.ERR, "Timeout writing to CAN sink socket")
            os.exit(-6)
        end

        self.buf.len = 0
        self.count = 0
    end,
    enqueue = function(self, msg)
        if self.count >= self.max_count then
            log:log(log.ERR, out_of_space_msg)
            os.exit(-7)
        elseif self.count >= (self.max_count / 2) then
            self:flush()
        end
        if self.base_timestamp == 0 then return end
        local dmsg = self.buf.msg + self.count
        dmsg.timestamp[0] = htonll(msg.timestamp[0] - self.base_timestamp)
        -- The ID's MUST be in network byte order
        ffi.cast('uint32_t *', msg.value)[0] =
            C.htonl(ffi.cast('uint32_t *', msg.value)[0])
        ffi.copy(dmsg.value, msg.value, 13)
        self.count = self.count + 1
        -- If this is a heartbeat, flush the buffer
        if bit.band(dmsg.value[12], 0x20) == 0x20 then
            self:flush()
        end
    end,
}})

----------------------------------------------------------------------

local can_proxy = class('can_proxy', channel_proxy)

function can_proxy:initialize(name, devname)
    self.proxy_type = 'can'
    self.devname = devname
    self.device = string.format(
        "CAN%d", tonumber(devname:match('/dev/cca(%d+)')) + 1)
    cca_handler.new(devname, self, can,
        'CAN_' .. tonumber(devname:match('/dev/cca(%d+)')) ..
        '.bin', sample_clock.master_rate)

    self:cmd_block('txdr_passthru_power_on')    -- Need to fix someday
    self.channel_sinks = {}
    channel_proxy.initialize(self, name)
    self.connector = '@' .. self.proxy_hostname
end

function can_proxy:data_unavailable()
    for w, _ in pairs(self.channel_sinks) do
        w.md_sink{
            error_message =
                "Proxy has stopped receiving data from CAN hardware"
        }
        w.cleanup(w)
        os.exit(-10)
    end
    os.exit(-12)
end

-- Function triggered by control API request to dump incoming
-- raw CAN messages to a file, for use by QA / Cal Stand
function can_proxy:buffer_to_log(msg)

    if not self.log_file then return end
    if msg.cmd ~= C.CAN_RX_START then return end

    local buf = ffi_cast('can_rxtx_buffer_t *', msg.bufdata)
    for i = 1, msg.element_count do

        local dlc = bit.band(buf.frame_format, 15)
        if dlc < 1 then return end

        -- A loop would clearly make more sense here but this
        -- unrolled code is faster and is performance-sensitive
        -- (have to be able to keep up with a high-rate CAN source)
        local str
        if dlc == 1 then
            str = string.format("%08X %02X\n",
                buf.can_id, buf.can_message[0])
        elseif dlc == 2 then
            str = string.format("%08X %02X %02X\n",
                buf.can_id, buf.can_message[0],
                buf.can_message[1])
        elseif dlc == 3 then
            str = string.format("%08X %02X %02X %02X\n",
                buf.can_id, buf.can_message[0],
                buf.can_message[1], buf.can_message[2])
        elseif dlc == 4 then
            str = string.format(
                "%08X %02X %02X %02X %02X\n",
                buf.can_id, buf.can_message[0],
                buf.can_message[1], buf.can_message[2],
                buf.can_message[3])
        elseif dlc == 5 then
            str = string.format(
                "%08X %02X %02X %02X %02X %02X\n",
                buf.can_id, buf.can_message[0],
                buf.can_message[1], buf.can_message[2],
                buf.can_message[3], buf.can_message[4])
        elseif dlc == 6 then
            str = string.format(
                "%08X %02X %02X %02X %02X %02X %02X\n",
                buf.can_id, buf.can_message[0],
                buf.can_message[1], buf.can_message[2],
                buf.can_message[3], buf.can_message[4],
                buf.can_message[5])
        elseif dlc == 7 then
            str = string.format(
                "%08X %02X %02X %02X %02X %02X %02X %02X\n",
                buf.can_id, buf.can_message[0],
                buf.can_message[1], buf.can_message[2],
                buf.can_message[3], buf.can_message[4],
                buf.can_message[5], buf.can_message[6])
        else
            str = string.format(
                "%08X %02X %02X %02X %02X %02X %02X %02X %02X\n",
                buf.can_id, buf.can_message[0],
                buf.can_message[1], buf.can_message[2],
                buf.can_message[3], buf.can_message[4],
                buf.can_message[5], buf.can_message[6],
                buf.can_message[7])
        end

        self.log_file:write(str)
        buf = buf + 1

    end

end

-- Function used to decode raw data from the FPGA
-- This is the 'real' buffer() function used for normal tests. It is
-- assigned as self.buffer at initialization. The logging control API
-- request redirects buffer() to buffer_to_log() above.
function can_proxy:buffer_to_sink(msg)
    if msg.cmd == C.CAN_RX_START then
        local buf = ffi_cast('can_rxtx_buffer_t *', msg.bufdata)
        for i = 1, msg.element_count do
            for _, sink in pairs(self.channel_sinks) do
                local m = ffi_cast(message_can.ptr_type, buf)
                local err = sink:enqueue(m)
                if err then
                    for w, sink in pairs(self.channel_sinks) do
                        w.md_sink{
                            error_message =
                                string.format("Error sending to test: %s", err),
                        }
                        w.cleanup(w)
                    end
                    return
                end
            end
            buf = buf + 1
        end
        if self.data_monitor:is_active() then
            self.data_monitor:again(ev.Loop.default)
        end
    elseif msg.cmd == C.CAN_EVENT_START then
        local buf = ffi_cast('can_ev_buffer_t *', msg.bufdata)
        for i = 1, msg.element_count do
            if buf.events.bus_off ~= 0 then
                for w, sink in pairs(self.channel_sinks) do
                    w.md_sink{error_message = 'CAN Bus is OFF'}
                    w.cleanup(w)
                end
                return
            end
            buf = buf + 1
        end
    end
end
can_proxy.buffer = can_proxy.buffer_to_sink

local spec_translate = {
    bit_rate = 'baud_rate',
}

-- Function used to start CAN data from the FPGA
function can_proxy:start_stream(w)
    w.proxy = self

    local device_spec = {}
    for k, v in pairs(self.spec) do
        device_spec[spec_translate[k] or k] = v
    end
    if not self.spec.internal_termination then
        self:cmd('disable_terminator')
    else
        self:cmd('enable_terminator')
    end
    self:cmd('bus_setup', device_spec)

    self:cmd('event_setup', {bus_off=true, bus_off_deassert=true})

    self:cmd_yield('clear_query_msgs')
    -- Uncomment next line to enable self reception of CAN messages.
    -- self:cmd_block('enable_self_rx')
    self:cmd_yield('rx_start', 0, update_rate, 16, 128)
    self:cmd_yield('event_start', 0, update_rate, 16, 128)

    if self.spec._active_queries then
        self:cmd_yield('add_query_msgs', self.spec._active_queries)
        self:cmd_yield('tx_start')
    end
end

-- Function used to stop the FPGA from sending CAN data
function can_proxy:stop_stream()
    self:cmd('event_stop')
    self:cmd('rx_stop')
    self:cmd('tx_stop')
    if self.spec and self.spec._active_queries then
        self:cmd('clear_query_msgs')
    end
end

-- Function used to assign a comet sink to channels
function can_proxy:start_comet_stream(w, channels)
end

-- Function called to start acquisiton on a channel
function can_proxy:start_channel_stream(w, spec, md_sink, sink)
    -- Don't allow more than one test to use this channel
    for _, _ in pairs(self.channel_sinks) do
        md_sink{
            error_message =
                "Channel can not be used in more than one test at a time",
        }
        return
    end
    -- Dont allow the control channel to be used in a test
    if spec.connector:find('^control') then
        md_sink{
            error_message = "Control channels can not be used in a test",
        }
        return
    end

    self:cmd_block('txdr_passthru_power_on')    -- Need to fix someday

    -- Initial channel setup
    if config.device_type ~= 'edaqxr' then
        self.start_stamp = self:cmd_block('get_current_ts').ts + 60000000
        -- Make it 1 second early
        self.start_stamp = self.start_stamp - sample_clock.master_rate
        spec.ts = self.start_stamp
        self.spec = spec

        -- Make sure the start time being supplied is doable!
        local cts = self:cmd_block('get_current_ts')
        if cts and cts.ts and cts.ts > self.start_stamp then
            md_sink{
                error_message = "Test start time in the past",
            }
            return
        end
    else
        self.spec = spec
    end

    -- Special 'sink' for raw CAN
    -- NOTE:  Back out start time adjustment so we get timestamps relative to
    -- the actual test start time.
    sink = ffi_new('can_sink', 0, w.socket, can_buf_count, 0)

    -- Insert into Channel Sink table
    self.channel_sinks[w] = sink
    
    self.start_callback = function(self, start_time)
            self.start_callback = nil
            -- For eDAQXR devices the sample clock generator is started by
            -- the test engine.  It is ALWAYS configured to start running
            -- exactly 1 second prior to the start time
            if config.device_type ~= 'edaqxr' then
                self.start_stamp = ts.ts_to_stamp(ts.iso_to_ts(start_time))
                sink.base_timestamp = self.start_stamp
                self.start_stamp = self.start_stamp - sample_clock.master_rate
            else
                sink.base_timestamp = 40000000
                armada.timing:SetStartTime(start_time,
                    spec.sample_rate_domain == 'binary' and 1.25 or 1.0)
            end
            self:launch_coroutine(function() self:start_stream(w) end)
            if self.data_monitor and not self.data_monitor:is_active() then
                self.data_monitor:start(ev.Loop.default)
            end
        end

    -- Cleanup function called when a test disconnects
    w.cleanup = function()
        self.channel_sinks[w] = nil
        for _, _ in pairs(self.channel_sinks) do return end
        if self.data_monitor and self.data_monitor:is_active() then
            self.data_monitor:stop(ev.Loop.default)
        end
        if config.device_type == 'edaqxr' then
            armada.timing:StopClock()
        end
        self:stop_stream()
        w.cleanup = nil
    end

    w.md_sink = md_sink
    self:announce_ready(spec.connector, true)
end

function can_proxy:available_channels(hostname)

    local chans = {}
    table.insert(chans, {
            name = 'raw_' .. hostname:gsub("%W", "_"),
            connector = '@' .. hostname,
            description = "Raw CAN messages",
            data_type = 'message_can',
            category = 'source/hbm_' ..  config.device_type .. '/msg_can',
            message_format = 'CAN',
        })
    table.insert(chans, {
            name = 'control_' .. hostname:gsub("%W", "_"),
            connector = 'control.status@' .. hostname,
            description = 'CAN Control Channel',
            data_type = 'message_raw',
            category = 'source/hbm_' .. config.device_type .. '/msg_can',
        })
    self.channels = chans
    return chans
end

function can_proxy:transducer_power(params)
    local msg
    local result
    if params[1] == 'on' then
        self:cmd_block('txdr_passthru_power_on')
        msg = 'Successfully turned CAN transducer power on'
        result = 'success'
    elseif params[1] == 'off' then
        self:cmd_block('txdr_passthru_power_off')
        msg = 'Successfully turned CAN transducer power off'
        result = 'success'
    else
        msg = 'Unknown parameter passed to CAN transducer power control API'
        result = 'error'
    end
    return {
        msg, result
    }
end

function can_proxy:termination_enable(params)
    local msg
    local result
    if params[1] == 'on' then
        self:cmd_block('enable_terminator')
        msg = 'Successfully enabled CAN termination'
        result = 'success'
    elseif params[1] == 'off' then
        msg = 'Successfully disabled CAN termination'
        result = 'success'
        self:cmd_block('disable_terminator')
    else
        msg = 'Unknown parameter passed to CAN termination control API'
        result = 'error'
    end
    return {
        msg, result
    }
end

function can_proxy:control(loop, w, http, spec)
    if not spec.request then
        return {
            msg = string.format('Control request not specified on %s',
                                self.proxy_hostname),
            result = 'error',
        }
    end
    if not spec.connector then
        return {
            msg = string.format('Control with no connector specified on %s',
                                self.proxy_hostname),
            result = 'error',
        }
    end
    if spec.request == 'start_log' then
        self.buffer = self.buffer_to_log
        self.log_file = io.open(string.format('/tmp/can%d.txt',
            1 + (self.devname:match('/dev/cca(%d+)') or -1)), 'w')
        self:launch_coroutine(function()
            self:cmd_block('txdr_passthru_power_on')
            if spec.params and spec.params.termination then
                self:cmd('enable_terminator')
            else
                self:cmd('disable_terminator')
            end
            self:cmd('bus_setup', { baud_rate = spec.params.bit_rate or 1000000 })
            self:cmd('event_setup', { bus_off = true, bus_off_deassert = true })
            self:cmd_yield('clear_query_msgs')
            self:cmd_yield('rx_start', 0, update_rate, 16, 128)
            self:cmd_yield('event_start', 0, update_rate, 16, 128)
            if config.device_type == 'edaqxr' then
                local start_time = ts.get_ts()
                start_time.s = start_time.s + 1
                local start_time_iso = ts.ts_to_iso(start_time)
                armada.timing:SetStartTime(start_time_iso, 1.0)
            end
        end)
        return {dumping = (self.log_file and true or false)}
    end
    if spec.request == 'stop_log' then
        if config.device_type == 'edaqxr' then
            armada.timing:StopClock()
        end
        self:cmd('event_stop')
        self:cmd('rx_stop')
        self:cmd('tx_stop')
        if self.log_file then self.log_file:close() end
        self.log_file = nil
        self.buffer = self.buffer_to_sink
        return {dumping = false}
    end
    for _, c in ipairs(self.channels) do
        if c.connector == spec.connector then
            if not self[spec.request] or
                type(self[spec.request]) ~= 'function' then
                return {
                    msg = string.format("Control request %s invalid on %s",
                                        spec.request, c.connector),
                    result = 'error',
                }
            end
            return self[spec.request](self, spec.params)
        end
    end
    return {
        msg = string.format("Control connector %s not found",
                            spec.connector),
        result = 'error',
    }
end

return can_proxy
