require 'strict'
require 'middleclass'
local sqlite3 = require 'lsqlite3'
local cjson = require 'cjson.safe'
local config = require 'system_config'
local network_db = require 'network_db'
local uuid = require 'uuid'
local ts = require 'timestamps'
local socket = require 'socket'
local smtp = require 'socket.smtp'
local log = require 'log'
local ffi = require 'ffi'
local C = ffi.C

ffi.cdef[[
    typedef int pid_t;

    pid_t fork(void);
    int close(int);
]]

local mail = class('mail')

--[[
The server id information in the setup database is a JSON object containing

{
    -- Ip address of the SMTP server.  This is parameter is required
    server_ip = "192.168.100.100",

    -- Domain name of the SMTP server.  This parameter is required
    domain_name = "gmail.com",

    -- Port number used to connect to the SMTP server.  This parameter is
    -- optional and will be defaulted to the appropriate values when needed.
    port = 25,

    -- Login name to authenticate to the SMTP server when using "login"
    -- authentication and/or TLS or STARTTLS security.  If "authentication"
    -- mode is set to "none" and TLS/STARTTLS are not being used, this
    -- parameter is not needed.
    username = "mailuser",

    -- Password to authenticate to the SMTP server when using "login"
    -- authentication and/or TLS or STARTTLS security.  If "authentication"
    -- mode is set to "none" and TLS/STARTTLS are not being used, this
    -- parameter is not needed.
    password = "password",

    -- Securtiy mechanism to use.  This paramater is optional and if not
    -- specified "security" defaults to "none".  Valid values for this
    -- parameter are "none", "tls", "starttls".
    security = "tls",

    -- The parameter below is optional.  If not specified it defaults to
    -- "none".  If "login" is specified, then "username" and "password" are
    -- required.  Valid values for this parameter are "login" and "none".
    authentication_method = "none",
}

]]

function mail:initialize(json)
    -- Set the default send function that does nothing
    self.send = function(self, mail) end

    -- Fake the server info if system configuration says to do so
    if config.use_fake_email_server_info then
        self.server_info = {
            server_ip = "webmail.hbm.com",
            domain_name = "hbm.com",
            port = 25,
            authentication_method = "none",
            username = "somebody",      -- Doc purposes only, not used
            password = "hbm12345",      -- Doc purposes only, not used
            security = "none",
        }
        self.send = self._send
        return
    end

    if not json then
        -- Look up the server information from the setup database.  If it is not
        -- found.  Thrown an error.
        local db = sqlite3.open('/var/lib/armada/setup.db')
        if not db then
            error("Unable to open setup database")
            return
        end
        for r in db:rows(string.format([[
            SELECT content FROM tsd_users WHERE db_name = 'mail_servers' AND
                deleted = '' AND content LIKE '%%"active":true%%';
            ]])) do
            self.server_info = cjson.decode(r[1])
            if not self.server_info then
                error("Mail:  Unable to decode email server record")
                db:close()
                return
            end
            break
        end
        db:close()
    else
        self.server_info = cjson.decode(json)
    end

    -- Make sure we found a server record
    if not self.server_info then
        -- No record found, no error, mail send function does nothing
--        error("Mail:  No active email server record found")
        return
    end

    if not self.server_info.server_ip or
        not self.server_info.server_ip:find('^%d+%.%d+%.%d+%.%d+$') then
        error("Mail:  Invalid or malformed SMTP server address")
        return
    end

    -- Use the real send function
    self.send = self._send
end

--[[
mail is a LUA table that looks like the following

{
    from = "youremail",
    subject = "Subjet of the email",
    to = "to.person1@mail.blah.com",
    cc = {"ccto.person@hotmail.com", "cc.someone@apamadax.com"},
    bcc = {"blindcc.someone@hbm.com", "blindcc2@gmail.com"},
    body = "This is the body of the email".
    log_lines = 200,
}

In the above table, "subject", "to", and "body" are required.  The "cc" and
"bcc" fields can contain more than one entry each.  The "from" field is not
required and will be defaulted to "devicename-deviceserialnumber@server_domain"
if not specified.  If specified, it should only be the part of email address
of the sender to the left of the "@" sign.  The log lines parameter is optional
and if included specifies the number of lines from the end of the system log
to include in the email.
]]

function mail:_send(mail)
    -- Make sure server info was found
    if not self.server_info then
        error("Mail no valid server information")
        return
    end

    -- Make sure the argument is the expected type
    if not mail or type(mail) ~= 'table' then
        error("Mail 'sendmail' function requires 'table' argument")
        return
    end

    -- Make sure the "required" information was given
    if not mail.subject or not mail.to or not mail.body or
        type(mail.subject) ~= 'string' or type(mail.to) ~= 'table' or
        type(mail.body) ~= 'string' then
        error(
            "Sending mail requires 'Subject', 'To', and 'Body' to be provided")
        return
    end

    for i, v in ipairs(mail.to or {}) do
        if not mail.to[i]:find('^<.*>$') then
            mail.to[i] = '<' .. v .. '>'
        end
    end
    for i, v in ipairs(mail.cc or {}) do
        if not mail.cc[i]:find('^<.*>$') then
            mail.cc[i] = '<' .. v .. '>'
        end
    end
    for i, v in ipairs(mail.bcc or {}) do
        if not mail.bcc[i]:find('^<.*>$') then
            mail.bcc[i] = '<' .. v .. '>'
        end
    end

    local from = mail.from or self.server_info.from
    if not from then
        local ndb = network_db:new()
        from = 'SoMatXR@' .. ndb:get_inside_network_hostname()
        ndb:close()
    end
    self.server_info.domain_name = from:match('^.+@(.+)$')

    local msg = {
        headers = {
            from = from,
            to = table.concat(mail.to, ','),
            cc = mail.cc and table.concat(mail.cc, ','),
            bcc = mail.bcc and table.concat(mail.bcc, ','),
            subject = mail.subject
        },
        body = string.format("Email generated at: %s (UTC system time)\n\n",
            ts.ts_to_iso(ts.get_ts())) .. mail.body
    }

    -- Add log lines if specified
    if mail.log_lines and tonumber(mail.log_lines) and
        tonumber(mail.log_lines) > 0 then
        msg.body = msg.body ..
            string.format("\n------------ Last %d log entries -------------\n",
                tonumber(mail.log_lines))

        local fh = io.popen('tail-log ' .. tostring(mail.log_lines))
        if fh then
            msg.body = msg.body .. fh:read('*all')
            fh:close()
        end
    end

    local rcpt = {}
    for _, v in ipairs(mail.to or {}) do rcpt[#rcpt + 1] = v end
    for _, v in ipairs(mail.cc or {}) do rcpt[#rcpt + 1] = v end
    for _, v in ipairs(mail.bcc or {}) do rcpt[#rcpt + 1] = v end

    local source = smtp.message(msg)
    local _email = {
        from = '<' .. from .. '>',
        rcpt = rcpt,
        source = source,
        user = self.server_info.username,
        password = self.server_info.password,
        server = self.server_info.server_ip,
        port = self.server_info.port,
        domain = self.server_info.domain_name,
    }

    log:log(log.INFO, string.format("Sending email\n%s", cjson.encode(msg)))

    local pid = C.fork()
    if pid == 0 then
        -- We are child
        local ev = require 'ev'
        ev.Loop.default:fork()
        for i = 3, 255 do C.close(i) end
        local ok, err = smtp.send(_email)
        if not ok then
            log:log(log.NOTICE, err)
        else
            log:log(log.INFO, "Email sent successfully")
        end
        os.exit(1)
    elseif pid > 0 then
        -- We are parent
        return true
    end

    -- error
    return nil, 'Unable to fork to send email'
end

function mail:_busybox_send(mail)
    -- Make sure server info was found
    if not self.server_info then
        error("Mail no valid server information")
        return
    end

    -- Make sure the argument is the expected type
    if not mail or type(mail) ~= 'table' then
        error("Mail 'sendmail' function requires 'table' argument")
        return
    end

    -- Make sure the "required" information was given
    if not mail.subject or not mail.to or not mail.body or
        type(mail.subject) ~= 'string' or type(mail.to) ~= 'table' or
        type(mail.body) ~= 'string' then
        error(
            "Sending mail requires 'Subject', 'To', and 'Body' to be provided")
        return
    end

    -- Start building the command
    local cmd = 'busybox sendmail -t -f '
    if not mail.from then
        local ndb = network_db:new()
        cmd = cmd .. (ndb:get_inside_network_hostname() or
            config.device_type)
        ndb:close()
    else
        cmd = cmd .. mail.from
    end
    cmd = cmd .. '@' .. self.server_info.domain_name

    -- Add the security stuff if needed
    -- FIXME:  The security stuff is untested because I don't have access to
    -- an SMTP server with TLS or STARTTLS.
    if self.server_info.security then
        if self.server_info.security == 'tls' then
            cmd = cmd .. " -H 'exec openssl s_client -quiet -tls1 " ..
                "-CApath /usr/ssl -connect " .. self.server_info.server_ip ..
                ":" ..  (tostring(self.server_info.port) or "465") .. "'"
        elseif self.server_info.security == 'starttls' then
            cmd = cmd .. " -H 'exec openssl s_client -quiet -tls1 -starttls " ..
                "-CApath /usr/ssl -connect " ..  self.server_info.server_ip ..
                ":" .. (tostring(self.server_info.port) or "587") .. "'"
        elseif self.server_info.security == 'none' then
            -- Nothing to do here
            cmd = cmd .. ' -S ' .. self.server_info.server_ip
            if self.server_info.port then
                cmd = cmd .. ':' .. tostring(self.server_info.port)
            end
        else
            error("Unknown security type '" .. self.server_info.security .. "'")
            return
        end
    else
        cmd = cmd .. ' -S ' .. self.server_info.server_ip
        if self.server_info.port then
            cmd = cmd .. ':' .. tostring(self.server_info.port)
        end
    end

    -- Add the server and port
    -- FIXME:  Not sure if this is needed with TLS or STARTTLS
--[[
    cmd = cmd .. ' -S ' .. self.server_info.server_ip
    if self.server_info.port then
        cmd = cmd .. ':' .. tostring(self.server_info.port)
    end
]]

    -- Set up for 'login' authentication method if configured to do so.  If
    -- not, assume no authentication is required.
    if self.server_info.authentication_method then
        if self.server_info.authentication_method == 'login' then
            if not self.server_info.username or
                not self.server_info.password then
                error("Login authentication method requires 'username' and " ..
                    "'password'")
                return
            end
            cmd = cmd .. " -au" .. self.server_info.username .. " -ap" ..
                self.server_info.password
        elseif self.server_info.authentication_method ~= 'none' then
            error("Unknown authentication method '" ..
                self.server_info.authentication_method .. "'")
            return
        end
    end

    cmd = cmd .. " " .. mail.to[1]

    -- Create the text file to be sent
    local mail_txt = '/tmp/' .. uuid.generate()
    local fh = io.open(mail_txt, 'w')
    if not fh then
        error("Mail:  Unable to create mail text file")
        return
    end

    -- Add additional "To" recipients
    for i, to in ipairs(mail.to) do
        if i > 1 then
            fh:write(string.format("To: %s\n", to))
        end
    end

    -- Add "Cc" recipients, if any
    if mail.cc and next(mail.cc) then
        for _, cc in ipairs(mail.cc) do
            fh:write(string.format("Cc: %s\n", cc))
        end
    end

    -- Add "Bcc" recipients, if any
    if mail.bcc and next(mail.bcc) then
        for _, bcc in ipairs(mail.bcc) do
            fh:write(string.format("Bcc: %s\n", bcc))
        end
    end

    -- Add subject
    fh:write("Subject: " .. mail.subject .. "\n\n")

    -- Generate body
    fh:write(string.format("Email generated at: %s (UTC system time)\n\n",
        ts.ts_to_iso(ts.get_ts())))
    fh:write(mail.body)

    -- Add log lines if specified
    if mail.log_lines and tonumber(mail.log_lines) and
        tonumber(mail.log_lines) > 0 then
        fh:write(
            string.format("\n------------ Last %d log entries -------------\n",
            tonumber(mail.log_lines)))
    end
    fh:close()
    if mail.log_lines and tonumber(mail.log_lines) and
        tonumber(mail.log_lines) > 0 then
        os.execute('tail-log ' .. tostring(mail.log_lines) .. " >>" ..
            mail_txt)
    end
    
    -- Finish up command to send email
    cmd = cmd .. " <" .. mail_txt

    cmd = cmd .. " 2>&1 | logger -t " .. arg[0] .. " -p user.warn"

    local f = '/tmp/' .. uuid.generate() .. '.sh'
    local fh = io.open(f, 'w')
    fh:write("#!/bin/sh\n\n")
    for i = 3, 255 do
        fh:write(string.format("exec %d<&-\n", i))
    end
    fh:write(cmd)
    fh:write("\nrm -f " .. mail_txt .. "\n")
    fh:write("rm -f " .. f .. "\n")
    fh:close()
    os.execute('chmod a+x ' .. f)

    -- Get it sent
    os.execute(f .. ' &')
end

return mail
