--! \file

require('lj_include')(
    'unistd',
    'signal'
)

require 'middleclass'
local ev = require 'ev'
local ffi = require 'ffi'
local C = ffi.C
local net = require 'net_ffi'
local cjson = require 'cjson'
local underscore = require 'underscore'
local trans = require 'mx.translator'
local config = require 'system_config'
local log = require 'log'
local teds_utils = require 'teds_utils'

require 'xpath'
require 'lxp/lom'

local xml_declaration = '<?xml version="1.0" encoding="UTF-8"?>\n'

local Configure = class('Configure')

local htonl = ffi.abi('le') and bit.bswap or bit.tobit

local htonll_union =
    ffi.typeof('union { unsigned long long u64; unsigned long u32[2]; }')
local function htonll(x)
    if ffi.abi('le') then
        local v = htonll_union(x)
        v.u32[0], v.u32[1] = bit.bswap(v.u32[1]), bit.bswap(v.u32[0])
        return v.u64
    end
    return x
end

local function htons(x)
    if ffi.abi('le') then
        return bit.rshift(htonl(x), 16)
    end
    return htonl(x)
end

local function gsub_literal(str, pattern, replacement)
    replacement = string.gsub(replacement, '%%', '%%%%')
    return string.gsub(str, pattern, replacement)
end

local function generate_mx_results(error_message, xml, ecode, all_errors)
    if ecode == C.ERR_SUCCESS then
        return {
            result = 'success',
            msg = 'Transaction successful'
        }
    else
        local msg = error_message .. ':\n'
        for _, err in ipairs(all_errors) do
            msg = msg .. string.format('%d (%s): %s\n', err.id, err.type, err.msg)
        end
        msg = msg .. '\nTransaction XML was:\n\n' .. xml
        return {
            result = 'error',
            msg = msg
        }
    end
end

local ntohl = htonl
local ntohs = htons
local ntohll = htonll

function Configure:_send_transaction_request(v1, v2, callback, data)
    local dlen = data and #data or 0

    local dump_fn
    local label = self:get_label()

    if config and config.dump_mx_transactions and data and #data > 0 then
        local name = self.devinfo.name or 'unknown'
        local serial = self.devinfo.serial_number or 'unknown'
        dump_fn = string.format('/tmp/mx-%s-%s-transactions.txt', name, serial)
        local fh = io.open(dump_fn, 'a')
        fh:write(string.format('@ %s, v1=%d, v2=%d, label=%d\n',
            os.date('%Y-%m-%d %H:%M:%S'), v1, v2, label))
        fh:write(data)
        fh:write('\n\n')
        io.close(fh)
    end

    -- Generate pointers to the assorted portions of the message
    local comm_hdr = ffi.cast('struct communications_hdr *', self.buf)
    local tr_hdr = ffi.cast('struct transaction_hdr *', comm_hdr + 1)
    local tr_req = ffi.cast('struct transaction_req *', tr_hdr + 1)
    local dptr = ffi.cast('char *', tr_req + 1)
    local comm_tail = ffi.cast('struct communications_tail *', dptr + dlen)

    -- Fill in communications header
    comm_hdr.start = htonl(C.START_OF_MESSAGE)
    comm_hdr.len = ffi.sizeof('struct communications_hdr')
    comm_hdr.next_hdr_type = C.NEXT_HDR_TYPE_TRANSACTION
    comm_hdr.payload_length = htons(ffi.sizeof('struct transaction_hdr') +
        ffi.sizeof('struct transaction_req') + dlen)
    comm_hdr.uuid = self.my_uuid

    -- Fill in transaction request header
    tr_hdr.type = htons(C.TRANSACTION_TYPE_REQUEST)
    tr_hdr.label = label
    tr_hdr.uuid = self.remote_uuid

    -- Fill in transaction request
    tr_req.data[0] = htons(v1)
    tr_req.data[1] = htons(v2)

    -- Fill in additional transaction data
    if dlen > 0 then
        ffi.copy(dptr, data, dlen)
    end

    -- Fill in communications tail
    comm_tail.tail = htonl(C.END_OF_MESSAGE)
    
    -- Set the callback function to be called when the transaction completes
    self.callback[tr_hdr.label] = function(_self, _error_code, _xml, _all_errors)
        if config and config.dump_mx_transactions and
            data and #data > 0 then
            local fh = io.open(dump_fn, 'a')
            if _all_errors and #_all_errors > 0 then
                fh:write(string.format('Errors for label %d:\n', label))
                for _, err in ipairs(_all_errors) do
                    fh:write(string.format('    %d (%s): %s\n', err.id, err.type, err.msg))
                end
                fh:write('\n')
            else
                fh:write(string.format('No errors for label %d.\n\n', label))
            end
            io.close(fh)
        end
        if callback then callback(_self, _error_code, _xml, _all_errors) end
    end

    -- Figure out how much data needs to be sent
    local len = ffi.sizeof('struct communications_hdr') +
        ffi.sizeof('struct transaction_hdr') +
        ffi.sizeof('struct transaction_req') +
        dlen + ffi.sizeof('struct communications_tail')

    -- Send it
    local sent = self.sbuf:write(self.buf, len)
    if sent ~= len then
    end
end

function Configure:_send_segment_ack(label)
    -- Generate pointers to the assorted portions of the message
    local comm_hdr = ffi.cast('struct communications_hdr *', self.buf)
    local tr_hdr = ffi.cast('struct transaction_hdr *', comm_hdr + 1)
    local tr_req = ffi.cast('struct transaction_req *', tr_hdr + 1)
    local comm_tail = ffi.cast('struct communications_tail *', tr_req + 1)

    -- Fill in communications header
    comm_hdr.start = htonl(C.START_OF_MESSAGE)
    comm_hdr.len = ffi.sizeof('struct communications_hdr')
    comm_hdr.next_hdr_type = C.NEXT_HDR_TYPE_TRANSACTION
    comm_hdr.payload_length = htons(ffi.sizeof('struct transaction_hdr') +
        ffi.sizeof('struct transaction_req'))
    comm_hdr.uuid = self.my_uuid

    -- Fill in transaction request header
    tr_hdr.type = htons(C.TRANSACTION_TYPE_RESPONSE_SEGMENT_ACK)
    tr_hdr.label = label
    tr_hdr.uuid = self.remote_uuid

    -- Fill in transaction request
    tr_req.data[0] = htons(0)
    tr_req.data[1] = htons(0)

    -- Fill in communications tail
    comm_tail.tail = htonl(C.END_OF_MESSAGE)
    
    -- Figure out how much data needs to be sent
    local len = ffi.sizeof('struct communications_hdr') +
        ffi.sizeof('struct transaction_hdr') +
        ffi.sizeof('struct transaction_req') +
        ffi.sizeof('struct communications_tail')

    -- Send it
    local sent = self.sbuf:write(self.buf, len)
    if sent ~= len then
    end
end

local warning_is_error = {
    ['119015'] = true,                  -- Sample rate not as requested
    ['119010'] = true,                  -- Filter frequency changed
    ['119011'] = true,                  -- Filter type changed
    ['119012'] = true,                  -- Filter characteristics changed
}

function Configure:_warning_is_error(id)
    return warning_is_error[tostring(id)]
end

-- Function used to process transaction response messages
function Configure:process_trans_response(buf, size)
    self.tr_responses = self.tr_responses or {}
    local comm_hdr = ffi.cast('struct communications_hdr *', buf)
    local tr_hdr = ffi.cast('struct transaction_hdr *', comm_hdr + 1)
    local tr_resp = ffi.cast('struct transaction_resp_hdr *', tr_hdr + 1)
    local len = size -
        (ffi.sizeof('struct communications_hdr') +
        ffi.sizeof('struct transaction_hdr') +
        ffi.sizeof('struct transaction_resp_hdr') +
        ffi.sizeof('struct communications_tail'))
    local xml
    if len > 0 then
        xml = ffi.string(ffi.cast('unsigned char *', tr_resp + 1), len)
    end
    local error_code = ntohl(tr_resp.error_code)
    local label = tonumber(tr_hdr.label)
    self.tr_responses[label] = {
        xml = xml or 'No information availble from MX device',
    }

    local all_errors = {}

    -- Essentially a no-op, but for unknown reasons this code avoids some
    -- non-deterministic and utterly broken LuaJIT behavior in this
    -- function. Chaos will ensue if this line of code is removed.
    if xml then print(xml) end

    if error_code ~= C.ERR_SUCCESS and xml then
        -- Get all of the messages from the returned XML
        local old_error_code = error_code
        error_code = C.ERR_SUCCESS
        local t = lxp.lom.parse(xml)
        if t then
            local msg = xpath.selectNodes(t, '/Reply/Message/Text/text()')
            local mtype = xpath.selectNodes(t, '/Reply/Message/Type/text()')
            local id = xpath.selectNodes(t, '/Reply/Message/Id/text()')
            if msg then
                for idx, m in ipairs(msg) do
                    all_errors[#all_errors + 1] = {
                        id = id[idx],
                        type = mtype[idx],
                        msg = msg[idx]
                    }
                    if mtype[idx] == 'warning' and
                        id[idx] and not self:_warning_is_error(id[idx]) then
                        -- Ignore all warnings except those about sample rate
                        -- changes because warnings may naturally occur when
                        -- a channel is being set up due to short term
                        -- inconsistancies between parameters during channel
                        -- reconfiguration.

                        -- Ignore warnings about output unit being disabled
                        -- because the physical units are set to "UserDefined"
                        if id[idx] ~= '104113' and id[idx] ~= '104112' and
                            -- Below is like this because I could not locate
                            -- the code for the message.
                            not m:find("physical engineering unit of sensor "..
                                "and amplifier's native unit are from the "..
                                "same class") then
                            self.logger:log(self.logger.WARNING, string.format(
                                "IGNORED WARNING #%s from MX Device: '%s'",
                                tostring(id[idx]), m))
                        end
                    else
                        error_code = old_error_code
                    end
                end
            end
        end
    end

    -- Set the error code to the possibly modified codd
    self.tr_responses[label]['error_code'] = error_code

    local cb = self.callback[label]
    if cb then
        self.callback[label] = nil
        cb(self, error_code, xml, all_errors)
    end
    self.tr_responses[label] = nil
end

-- Function used to process transaction reponse pending messages
function Configure:process_trans_response_pending(buf, size)
    -- Nothing needs to be done here
end

-- Function used to process transaction segment init messages
function Configure:process_trans_request_segment_init(buf, size)
    local comm_hdr = ffi.cast('struct communications_hdr *', buf)
    local tr_hdr = ffi.cast('struct transaction_hdr *', comm_hdr + 1)
    local sih = ffi.cast('struct segment_init_hdr *', tr_hdr + 1)
    local label = tr_hdr.label

    self.segmented_response = self.segmented_response or {}
    if self.segmented_response[tr_hdr.label] then
        self.logger:log(self.logger.ERR, string.format(
            "More than one outstanding segmented response in process for %d",
            tr_hdr.label))
        if cb then
            cb(self, C.ERR_NO_SUCCESS, nil)
        end
        return
    end
    local total_expected = ntohl(sih.total_data) - 8
    self.segmented_response[label] = {
        total_expected = total_expected,
        data = ffi.new('unsigned char[?]', total_expected),
    }
    self:_send_segment_ack(label)
end

-- Function used to process transaction segment messages
function Configure:process_trans_request_segment(buf, size)
    local comm_hdr = ffi.cast('struct communications_hdr *', buf)
    local tr_hdr = ffi.cast('struct transaction_hdr *', comm_hdr + 1)
    local label = tr_hdr.label
    local cb = self.callback[label]

    local sri = self.segmented_response[label]

    if not sri then
        self.logger:log(self.logger.ERR, string.format(
            "Unexpected segment from %d", label))
        self.segmented_response[label] = nil
        if cb then
            self.callback[label] = nil
            cb(self, C.ERR_NO_SUCCESS, nil)
        end
        return
    end

    local data = ffi.cast('unsigned char *', tr_hdr + 1)
    local dlen = ntohs(comm_hdr.payload_length) - 12
    if not sri.total_received then
        local trh = ffi.cast('struct transaction_resp_hdr *', tr_hdr + 1)
        local error_code = ntohl(trh.error_code)
        if error_code ~= C.ERR_SUCCESS then
            self.logger:log(self.logger.ERR, string.format(
                "Error in segment response: %d", error_code))
            self.segmented_response[label] = nil
            if cb then
                self.callback[label] = nil
                cb(self, error_code, nil)
            end
            return
        end
        data = ffi.cast('unsigned char *', trh + 1)
        dlen = ntohs(comm_hdr.payload_length) - 20
        sri.total_received = 0
    end
    ffi.copy(sri.data + sri.total_received, data, dlen)
    sri.total_received = sri.total_received + dlen

    self:_send_segment_ack(label)

    if sri.total_expected <= sri.total_received then
        if cb then
            self.callback[label] = nil
            local xml = ffi.string(ffi.cast('unsigned char *', sri.data),
                sri.total_received)
            cb(self, C.ERR_SUCCESS, xml)
        end
        self.segmented_response[label] = nil
        self.tr_responses[label] = nil
    end
end

-- Function used to process transaction segment pending messages
function Configure:process_trans_response_segment_pending(buf, size)
    -- Nothing needs to be done here
end

Configure.process_trans_message = {}

Configure.process_trans_message[C.TRANSACTION_TYPE_RESPONSE] =
    Configure.process_trans_response
Configure.process_trans_message[C.TRANSACTION_TYPE_RESPONSE_PENDING] =
    Configure.process_trans_response_pending
Configure.process_trans_message[C.TRANSACTION_TYPE_RESPONSE_SEGMENT_INIT] =
    Configure.process_trans_request_segment_init
Configure.process_trans_message[C.TRANSACTION_TYPE_RESPONSE_SEGMENT] =
    Configure.process_trans_request_segment
Configure.process_trans_message[C.TRANSACTION_TYPE_RESPONSE_SEGMENT_PENDING] =
    Configure.process_trans_response_segment_pending

-- Function used to process tramsaction messages
function Configure:_process_transaction(buf, size)
    local comm_hdr = ffi.cast('struct communications_hdr *', buf)
    local tr_hdr = ffi.cast('struct transaction_hdr *', comm_hdr + 1)
    local process_message = self.process_trans_message[ntohs(tr_hdr.type)]
    if not process_message then
        self.logger:log(self.logger.ERR,
            string.format("Unknown transaction type in response: %d",
            ntohs(tr_hdr.type)))
        return
    end
    process_message(self, buf, size)
    return size
end


local get_param_xml_fmt = xml_declaration ..
    '<SrvParams>' ..
        '<View>%d</View>' ..
        '<CurrChoicesOnly>0</CurrChoicesOnly>' ..
    '</SrvParams>'

function Configure:_get_view_tag_value(tag)
    for _, v in ipairs(self.view_list_parsed[self.current_view_num]) do
        if v.tag and v.tag == tag then return v[1] end
    end
end

function Configure:_get_view_info(first)
    if first then
        self.next_view_num = 1
    end
    while self.next_view_num < 100 do
        self.current_view_num = self.next_view_num
        self.next_view_num = self.next_view_num + 1
        if self.view_list_parsed[self.current_view_num] then
            local id = self:_get_view_tag_value('Id')
            local desc = self:_get_view_tag_value('Description')
            if id == '5' and desc ~= "Analog input connector parameters" then 
            	desc = "Analog input connector parameters"
            end
            if id and desc then return id, desc end
        end
    end
end

-- Function used to generate the JSON stuff used by a channel proxy
function Configure:_make_channel_info()
    self.channel_info = {}
    if self.device then
        for i = 1, self.num_channels do
            self.channel_info[#self.channel_info + 1] = self:get_spec(i)
        end
    end
    self:_normal_power_led()
    if self.configuration_finished_cb then
        self.configuration_finished_cb()
    end
end

-- Function used to flatten spec tables and remove JSON nulls
local function flatten(t, base)
    base = base or ''
    local r = {}
    for k, v in pairs(t) do
        if(type(v) == "table") then
            for k, v in pairs(flatten(v, k .. '_')) do
                if v ~= cjson.null then
                    r[base .. k] = v
                end
            end
        elseif v ~= cjson.null then
            r[base .. k] = v
        end
    end
    return r
end


-- Function used to create a "spec" table populated with information from the
-- XML of the specified channel.
function Configure:_get_channel_spec(cnum, st, ft, t, view, name)
--for k, v in pairs(t) do print(k, v) end
    local result = {}
    name = name or ''
    for k, path in pairs(t) do
        if type(path) == 'string' then
            path = gsub_literal(path, '#channel_num#', cnum)
            path = gsub_literal(path, '#sensor_input_mode#', st)
            path = gsub_literal(path, '#filter_type#', ft)
            if not path:find("@choice$") then
                path = path .. '/text()'
            end
            local v = xpath.selectNodes(view, path)[1]
--print("translator check", name .. k)
            local translator = self.parameter_translator[name .. k]
            if translator and v then
                v = translator(self, v, true)
            end
            result[k] = v
        elseif type(path) == 'table' then
            local n = string.format("%s%s_", name or '', tostring(k))
            result[k] =
                self:_get_channel_spec(cnum, st, ft, path, view, n)
        end
    end
    return result
end

-- Function used to convert a spec to internal format
function Configure:_convert_spec_to_internal(spec)

    -- Function used to return a calibration stanza by purpose name.
    local function get_cal_stanza(t, id)
        if not t then return end
        for _, v in ipairs(t) do
            if v.purpose == id and v.applied then
                return v
            end
        end
    end

    -- Get source units
    local define_source_units =
        get_cal_stanza(spec.calibration, 'define_source_units')
    if not define_source_units then
        define_source_units = {
            applied = true,
            purpose = 'define_source_units',
        }
        if not get_cal_stanza(spec.calibration, 'units_conversion') then
            define_source_units.units = spec.electrical_units
        elseif spec.source_physical_units or spec.electrical_units then
            define_source_units.units = spec.source_physical_units or
                spec.electrical_units
            define_source_units.text = define_source_units.units
        elseif spec.sensor_input and spec.sensor_input.mode then
            local e_unit_map = {
                volt = 'V',
                pt100 = '°C',
                pt1000 = '°C',
                thermocouple = '°C',
                bridge = 'mV/V',
                ['4-20ma'] = 'mA',
                iepe = 'A',
                resistor = 'Ohm',
                potentiometer = 'Ohm',
                ssi_encoder_counter = 'Imp',
            }
            define_source_units.units = 
                e_unit_map[spec.sensor_input.mode] or 'V'
            define_source_units.text = define_source_units.units
        end
    end

    -- Get all the calibration stanza's in the order needed for internal
    -- processing.  All unrecognized stanza's are ignored.
    local lead_wire_correction =
        get_cal_stanza(spec.calibration, 'lead_wire_correction')
    local sensor_scaling =
        get_cal_stanza(spec.calibration, 'sensor_scaling')

    if sensor_scaling and define_source_units and define_source_units.units then
        sensor_scaling.units = define_source_units.units
    end

    if lead_wire_correction and sensor_scaling and
        not lead_wire_correction.units and sensor_scaling.units then
        lead_wire_correction.units = sensor_scaling.units
    end

    -- Sensor scaling stanza.  Overridden by lead wire correction stanza if it
    -- exists.  If neither it or sensor scaling stanza exist, create one.
    sensor_scaling = lead_wire_correction or sensor_scaling or
        {
            purpose='sensor_scaling',
            mode = 'defined_slope_intercept',
            slope = '1',
            intercept = '0',
            -- FIXME:  Need to do something with units here but it is based
            -- on sensor type!  Same for mode!
        }

    -- If the sensor scaling mode is 'shunt' change it to 'defined_zero_span'
    if sensor_scaling.mode == 'shunt' then
        sensor_scaling.mode = 'defined_zero_span'
    end

    -- If sensor scaling mode is experimental two point, make it defined two
    -- point.
    if sensor_scaling.mode == 'experimental_two_point' then
        sensor_scaling.mode = 'defined_two_point'
    end
    -- If the sensor scaling mode is defined two point, make sure the number
    -- of points is specified
    if sensor_scaling.mode == 'defined_two_point' then
        sensor_scaling.num_points = 2
    end

    -- Units conversion stanza.  If there isn't one, create it based on info
    -- from the scaling stanza.
    local units_conversion =
        get_cal_stanza(spec.calibration, 'units_conversion') or
            {
                units = spec.electrical_units,
                mode = 'defined_slope_intercept',
                slope = '1',
                intercept = '0',
                purpose = 'units_conversion',
                text = spec.output_units,
            }
    if not units_conversion.units then
        units_conversion.units = spec.output_units
    end

    -- Zero balancing stanza.  If there isn't one, create it based on info from
    -- the units conversion stanza.
    local rezero = get_cal_stanza(spec.calibration, 'rezero') or
        {
            units = units_conversion.units,
            mode = 'not_applicable',
            slope = '1',
            intercept = '0',
            purpose = 'rezero',
        }
    if not rezero.units then
        rezero.units = units_conversion.units
    end
    if rezero.mode ~= 'not_applicable' then
        rezero.mode = 'all_test_runs'
    end
    rezero.intercept = rezero.intercept * -1

    spec.output_units = nil

    -- Now get everything in the proper place
    if spec.calibration and define_source_units and
        sensor_scaling and units_conversion and rezero then
        spec.calibration = {
            define_source_units = define_source_units,
            sensor_scaling = sensor_scaling,
            units_conversion = units_conversion,
            rezero = rezero,
        }
    end

    -- Override the sample rate with the legacy uncorrected rate
    if spec.sample_rate_legacy_uncorrected then
        spec.sample_rate = spec.sample_rate_legacy_uncorrected
    end

    -- Check to make sure filter configuration is not disabled.  Filter config
    -- is disabled by the 'set_channel_teds_enable' function to prevent the
    -- filter from getting turned off after the channel has been configured
    -- when a test is to be run.
    if not spec._disable_filter_config then
        -- Make sure digital filter type is set
        spec.digital_filter_type = spec.digital_filter_type or 'none'

        -- Set the digital filter choice based on the filter type
        spec._digital_filter_choice =
            (spec.digital_filter_type == 'none') and 'Off' or 'Lowpass'

        -- The the digital filter is off, don't set filter type or frequency
        if spec.digital_filter_type == 'none' then
            spec.digital_filter_type = nil
            spec.digital_filter_attenuation_frequency = nil
        end
    end

    -- Do any device specific massaging of the spec
    if self.device.to_internal then
        spec = self.device.to_internal(self, spec)
    end

    return spec
end

function Configure:_add_TEDS_units_to_spec(spec)
    local spec = spec
    local mode 
    local si = spec.sensor_input
    if si then
        mode = si.mode
    else
        return spec
    end
    for _, v in ipairs(spec.calibration) do
        if v.purpose and v.purpose == 'sensor_scaling' and v.units then
            spec.output_units = v.units
            spec.source_physical_units = v.units
            return spec
        end
    end
    local device = self.devinfo.name
        if device == "mx1615b" then
            if mode == "bridge" or mode == "potentiometer" then
                spec.output_units = "µm/m"
                spec.source_physical_units = "µm/m"
            elseif mode == "pt100" then
                spec.output_units = "°C"
                spec.source_physical_units = "°C"
            elseif mode == "resistor" then
                spec.output_units = "Ohm"
                spec.source_physical_units = "Ohm"
            else
                spec.output_units = "V"
                spec.physical_units = "V"
            end
        elseif device == "mx1601b" then
            if mode == "4-20ma" then
                spec.output_units = "mA"
                spec.source_physical_units = "mA"
            else
                spec.output_units = "V"
                spec.source_physical_units = "V"
            end
        elseif device == "mx1609b" then
            spec.output_units = "°C"
            spec.source_physical_units = "°C"
        elseif device == "mx411b" then
            if mode == "4-20ma" then
                spec.output_units = "mA"
                spec.source_physical_units = "mA"
            elseif mode == "bridge" or mode == "inductive_bridge" then
                spec.output_units = "µm/m"
                spec.source_physical_units = "µm/m"
            else
                spec.output_units = "V"
                spec.source_physical_units = "V"
            end
        elseif device == "mx840b" then
            if mode == "bridge" or mode == "inductive_bridge" 
                or mode == "potentiometer" or mode == "lvdt" then
                spec.output_units = "µm/m"
                spec.source_physical_units = "µm/m"
            elseif mode == "volt" or mode == "iepe" then 
                spec.output_units = "V"
                spec.source_physical_units = "V"
            elseif mode == "4-20ma" then
                spec.output_units = "mA"
                spec.source_physical_units = "mA"
            elseif mode == "pt100" or mode == "pt1000" or mode == "thermocouple" then
                spec.output_units = "°C"
                spec.source_physical_units = "°C"
            elseif mode == "resistor" then
                spec.sensor.output_units = "Ohm"
                spec.source_physical_units = "Ohm"
            elseif mode == "pulse_frequency" or mode == "encoder_frequency" then
                spec.output_units = "kHz"
                spec.source_physical_units = "kHz"
            elseif mode == "pulse_counter" or mode == "encoder_counter" then
                spec.output_units = "kImp"
                spec.source_physical_units = "kImp"
            elseif mode == "ssi_encoder_counter" then
                spec.output_units = "Imp"
                spec.source_physical_units = "Imp"
            end
        end 
    return spec
end

-- Deep copy a table
-- Code is from GitHub MihailJP/d_copy.lua
local function clone(t)
    if type(t) ~= "table" then return t end
    local meta = getmetatable(t)
    local target = {}
    for k, v in pairs(t) do
        if type(v) == "table" then
            target[k] = clone(v)
        else
            target[k] = v
        end
    end
    setmetatable(target, meta)
    return target
end

-- Function used to convert a spec to external format
function Configure:_convert_spec_to_external(spec)
    -- Make sure rezero cabibration step has everything.  Some of the info in
    -- this calibration step seems to get lost with TEDS sensors, so we
    -- restore it here.
    if spec.calibration.rezero and spec.calibration.rezero.intercept then
        spec.calibration.rezero.mode =
            spec.calibration.rezero.mode or 'defined_slope_intercept'
        spec.calibration.rezero.purpose =
            spec.calibration.rezero.purpose or 'rezero'
        spec.calibration.rezero.slope =
            spec.calibration.rezero.slope or 1
    end

    local spec = spec

    -- If the digital filter choice is set to "Off" then set the digital
    -- filter type to none.  Other filter parameters are as retreived from the
    -- device.
    if spec._digital_filter_choice then
        if spec._digital_filter_choice == 'Off' then
            spec.digital_filter_type = 'none'
        end
        spec._digital_filter_choice = nil
    end

    -- Sample rate drival
    spec.sample_rate = string.format("%.0f", tonumber(spec.sample_rate or 10))
    -- Do any device specific massaging of the spec
    if self.device.to_external then
        spec = self.device.to_external(self, spec)
    end

    -- If the units specified in sensor scaling are not set to user defined
    -- return those units, otherwise return the string from the user defined
    -- units.
    if spec.calibration.sensor_scaling.units == 'UserDefined' then
        spec.calibration.sensor_scaling.units =
            spec.calibration.sensor_scaling._units
    end
    spec.calibration.sensor_scaling._units = nil

    -- If the units specified in the output unit are not set to user defined
    -- return those units, otherwise return the string from the user defined
    -- units.
--[[
    if spec.calibration.units_conversion then
        if spec.calibration.units_conversion.units == 'UserDefined' then
            spec.calibration.units_conversion.units =
                spec.calibration.units_conversion._units
        end
        spec.calibration.units_conversion._units = nil
    end
]]
    spec.calibration.units_conversion = nil


    -- FIXME:  backward compatability, remove sometime
    spec.electrical_units = spec.calibration.sensor_scaling.units
    --spec.output_units = spec.calibration.units_conversion.units
-- FIXME:  OVERRIDE units with sensor scaling units
--spec.output_units = spec.calibration.sensor_scaling.units
    -- Remove to here

    -- For non-TEDS sensors or TEDS sensors using defined slope intercept mode
    -- remove all parameters not associated with defined slope intercept mode
    if spec.teds_sensor and not spec.teds_sensor.bin_data or
        spec.calibration.sensor_scaling.mode == 'defined_slope_intercept' then
        spec.calibration.sensor_scaling.physical_span = nil
        spec.calibration.sensor_scaling.electrical_span = nil
        spec.calibration.sensor_scaling.electrical_zero = nil
        spec.calibration.sensor_scaling.electrical = nil
        spec.calibration.sensor_scaling.physical = nil
        spec.calibration.sensor_scaling.num_points = nil
        spec.calibration.sensor_scaling.gage_factor = nil
        spec.calibration.sensor_scaling.bridge_factor = nil
        spec.calibration.sensor_scaling.purpose = 'sensor_scaling'
    -- For TEDS sensors using defined two point mode, remove all parameters not
    -- associated with defined two point mode
    elseif spec.calibration.sensor_scaling.mode == 'defined_two_point' then
        spec.calibration.sensor_scaling.physical_span = nil
        spec.calibration.sensor_scaling.electrical_span = nil
        spec.calibration.sensor_scaling.electrical_zero = nil
        spec.calibration.sensor_scaling.purpose = 'sensor_scaling'
    -- For TEDS sensors using defined zero span mode, remove all parameters not
    -- associated with defined zero span mode
    elseif spec.calibration.sensor_scaling.mode == 'defined_zero_span' then
        spec.calibration.sensor_scaling.electrical = nil
        spec.calibration.sensor_scaling.physical = nil
        spec.calibration.sensor_scaling.num_points = nil
        spec.calibration.sensor_scaling.purpose = 'sensor_scaling'
    elseif spec.calibration.sensor_scaling.mode == 'strain_gage' then
        spec.calibration.sensor_scaling.electrical = nil
        spec.calibration.sensor_scaling.physical = nil
        spec.calibration.sensor_scaling.purpose = 'sensor_scaling'
    end

    if spec.calibration.rezero and spec.calibration.rezero.intercept then
        spec.calibration.rezero.intercept =
            tonumber(spec.calibration.rezero.intercept) ~= 0 and
            (tonumber(spec.calibration.rezero.intercept) * -1) or 0
    end

    if spec.calibration.define_source_units then
        spec.calibration.define_source_units.purpose = 'define_source_units'
    end
    local cal
    if spec.calibration.units_conversion then
        spec.calibration.units_conversion.purpose = 'units_conversion'
    
    -- FIXME:  When we start using the items commented out below, enable them
        cal = {
--            spec.calibration.define_source_units,
            spec.calibration.sensor_scaling,
            spec.calibration.units_conversion,
            spec.calibration.rezero,
        }
    else 
    -- FIXME:  When we start using the items commented out below, enable them
        cal = {
--           spec.calibration.define_source_units,
            spec.calibration.sensor_scaling,
--          spec.calibration.units_conversion,
            spec.calibration.rezero,
        }
    end
    spec.calibration = cal
    spec = self:_add_TEDS_units_to_spec(spec)

    -- Copy all of the setup information to the TEDS sensor section so the
    -- GUI will have an "unmodified" data set to reference for TEDS "edited"
    -- determination.
    if spec.teds_sensor and spec.teds_sensor.use_teds and
        spec.teds_sensor.bin_data and spec.teds_sensor.bin_data ~= "" then

        spec.teds_sensor.calibration = clone(spec.calibration)
        spec.teds_sensor.sensor_input = clone(spec.sensor_input)
        spec.teds_sensor.source_physical_units = spec.source_physical_units
        spec.teds_sensor.output_units = spec.output_units
        spec.teds_sensor.source = 'teds'
        spec.teds_sensor.transducer_power = spec.transducer_power
        spec.teds_sensor.internal_termintation = spec.internal_termination
        if spec.special_parameters then
            spec.teds_sensor.special_parameters =
                clone(spec.special_parameters)
        end
    end

    return spec
end

-- Function used to supplement one table with entries in another table that
-- don't exist in the first table.
local function supplement_table(_to, from)
    local to = {}
    underscore.extend(to, _to)
    for k, v in pairs(from) do
        if type(v) == 'table' then
            to[k] = to[k] or {}
            to[k] = supplement_table(to[k], v)
        elseif not to[k] then
            to[k] = v
        end
    end
    return to
end

function Configure:_get_channel_info(cnum)
    -- Hardware version info
    local v = self.view['System parameters']
    local module = xpath.selectNodes(v, '//ModuleType/text()')[1]

    if cnum > 0 then
        local ver_array, sn_array = {}, {}
        if module:lower():find('mx471') then
            ver_array = xpath.selectNodes(v, '//Can/HardwareRevision/text()')
            sn_array = xpath.selectNodes(v, '//Can/SerialNumber/text()')
        elseif module:lower():find('mx878') then
            ver_array = xpath.selectNodes(v,
                '//AnalogOut/HardwareRevision/text()')
            sn_array = xpath.selectNodes(v,
                '//AnalogOut/SerialNumber/text()')
        else
            ver_array = xpath.selectNodes(v,
                '//AnalogIn/HardwareRevision/text()')
            sn_array = xpath.selectNodes(v,
                '//AnalogIn/SerialNumber/text()')
        end
        return {
            hw_revision = ver_array[cnum] or ver_array[1],
            serial_number = sn_array[cnum] or sn_array[1],
            connector = cnum,
        }
    end

    local sn = xpath.selectNodes(v, '//Platform/SerialNo/text()')[1]
    local cx_name = self.devinfo.name .. '-' .. sn:sub(-6, -1):lower()
    local alias = xpath.selectNodes(v, '//System/ModuleName/text()')[1]
    if alias then alias = alias:gsub('[^-a-zA-Z0-9 _]', '_') end
    if alias then alias = string.rep('_', (alias:match('^ +') or ''):len()) .. alias:gsub('^ +', '') end
    if alias and alias == cx_name then alias = '_' .. cx_name end
    if not alias or #alias == 0 then alias = xpath.selectNodes(v, '//System/ModuleLabel')[1] end

    return {
        alias = alias,
        device = xpath.selectNodes(v,
            '//System/ModuleLabel/text()')[1],
        hw_revision = xpath.selectNodes(v,
            '//Platform/HardwareRevision/text()')[1],
        fw_revision = xpath.selectNodes(v,
            '//Platform/FirmwareVersion/text()')[1],
        serial_number = xpath.selectNodes(v,
            '//Platform/SerialNo/text()')[1],
        housing = xpath.selectNodes(v,
            '//System/Housing/text()')[1],
    }
end

-- Function used to generate JSON compatible with channel configuration from
-- the current channel configuration information XML
function Configure:get_teds_spec(cnum)
    local view = self.view['Analog input connector parameters']

    -- Need the sensor type
    local stp = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/@choice'
    stp = gsub_literal(stp, '#channel_num#', cnum)
    local sensor_type = xpath.selectNodes(view, stp)[1]
    if not sensor_type then
        self.logger:log(self.logger.ERR, string.format(
            "Unable to get sensor type for %s channel %d", self.devinfo.name,
            cnum))
        return
    end

    -- Need the filter type
    local ftp = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/SyncSignal[@address="1"]/FilterType/@choice'
    ftp = gsub_literal(ftp, '#channel_num#', cnum)
    local filter_type = xpath.selectNodes(view, ftp)[1]
    if not filter_type then
        self.logger:log(self.logger.ERR, string.format(
            "Unable to get filter type for %s channel %d", self.devinfo.name,
            cnum))
        return
    end

    -- Actual channel configuration
    local spec = self:_get_channel_spec(cnum, sensor_type, filter_type,
        self.device.json2xpath, view)

    -- Return actual configuation
    return self:_convert_spec_to_external(spec)
end

-- Function used to generate JSON compatible with channel configuration from
-- the current and default channel configuration information XML
function Configure:get_spec(cnum)
    if not self.view['Analog input connector parameters'] then
        -- This is for MX878
        return supplement_table({}, self.device.default_spec)
    end
    local view = self.view['Analog input connector parameters']
    -- Need the sensor type
    local stp = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/@choice'
    stp = gsub_literal(stp, '#channel_num#', cnum)
    local sensor_type = xpath.selectNodes(view, stp)[1]
    if not sensor_type then
        self.logger:log(self.logger.ERR, string.format(
            "Unable to get sensor type for %s channel %d", self.devinfo.name,
            cnum))
        return
    end

    -- Need the filter type
    local ftp = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/SyncSignal[@address="1"]/FilterType/@choice'
    ftp = gsub_literal(ftp, '#channel_num#', cnum)
    local filter_type = xpath.selectNodes(view, ftp)[1]
    if not filter_type then
        self.logger:log(self.logger.ERR, string.format(
            "Unable to get filter type for %s channel %d", self.devinfo.name,
            cnum))
        return
    end

    -- Actual channel configuration
    local spec = self:_get_channel_spec(cnum, sensor_type, filter_type,
        self.device.json2xpath, view)

    -- Supplement default channel configuration with parameters that exist in
    -- actual channel configuration that don't exist in default channel config
    local def_spec = supplement_table(self.device.default_spec, spec)

    return self:_convert_spec_to_external(def_spec)
end

Configure._setup = {}

Configure.signal2_templates = {
    enabled = [[
                    <SyncSignal address="2">
                        <Output>
                            <RtAvailable>1</RtAvailable>
                        </Output>
                        <OutputRate>#signal2_sample_rate#</OutputRate>
#signal2_filter#
                    </SyncSignal>
]],
    disabled = [[
                    <SyncSignal address="2">
                        <Output>
                            <RtAvailable>0</RtAvailable>
                        </Output>
                    </SyncSignal>
]]
}

Configure.sensor_templates = {
    Bridge = [[
<?xml version="1.0" encoding="UTF-8"?>
    <Amplifier>
        <Connector#channel_num#>
            <Mode choice="AnalogInChannel">
                <AnalogInChannel>
                    <Sensor>
                        <Sensortype choice="Bridge">
                            <Bridge>
                                <MaxElectrical>#sensor_input_max#</MaxElectrical>
                                <BridgeType>#_bridge_type#</BridgeType>
                                <Wiring>#_bridge_wiring#</Wiring>
                                <Impedance>#bridge_resistance#</Impedance>
                                <ExciteAmpl>#excitation_range#</ExciteAmpl>
                                <ExciteFreq>#excitation_mode#</ExciteFreq>
                            </Bridge>
                        </Sensortype>
#sensor_scaling#
#sensor_units_template#
                    </Sensor>
                    <SyncSignal address="1">
                        <OutputRate>#sample_rate#</OutputRate>
#filter_template#
                        <Output>
                            <RtAvailable>#rt_enable#</RtAvailable>
                        </Output>
                    </SyncSignal>
#signal2#
                    <OutputUnit>
#units_template#
                    </OutputUnit>
                    <ZeroOffset>
                        <Inhibit>0</Inhibit>
                        <Offset>#calibration_rezero_intercept#</Offset>
                        <Target>0.000000</Target>
                    </ZeroOffset>
                </AnalogInChannel>
            </Mode>
        </Connector#channel_num#>
    </Amplifier>
]],
    Pt = [[
<?xml version="1.0" encoding="UTF-8"?>
    <Amplifier>
        <Connector#channel_num#>
            <Mode choice="AnalogInChannel">
                <AnalogInChannel>
                    <Sensor>
                        <Sensortype choice="Pt">
                            <Pt>
                                <PtType>#_pt_type#</PtType>
                                <Wiring>#_pt_wiring#</Wiring>
                            </Pt>
                        </Sensortype>
                        <Scaling>
                            <Scalingtype choice="Internal">
                                <Internal/>
                            </Scalingtype>
                        </Scaling>
#sensor_units_template#
                    </Sensor>
                    <SyncSignal address="1">
                        <OutputRate>#sample_rate#</OutputRate>
#filter_template#
                        <Output>
                            <RtAvailable>#rt_enable#</RtAvailable>
                        </Output>
                    </SyncSignal>
#signal2#
                    <OutputUnit>
#units_template#
                    </OutputUnit>
                    <ZeroOffset>
                        <Inhibit>0</Inhibit>
                        <Offset>#calibration_rezero_intercept#</Offset>
                        <Target>0.000000</Target>
                    </ZeroOffset>
                </AnalogInChannel>
            </Mode>
        </Connector#channel_num#>
    </Amplifier>
]],
    -- The units for this sensor mode are fixed at mV/V (1001)
    -- The "Resitance" parameter is left as is
    Potentiometer = [[
<?xml version="1.0" encoding="UTF-8"?>
    <Amplifier>
        <Connector#channel_num#>
            <Mode choice="AnalogInChannel">
                <AnalogInChannel>
                    <Sensor>
                        <Sensortype choice="Potentiometer">
                            <Potentiometer>
                                <Wiring>#_pot_wiring#</Wiring>
                            </Potentiometer>
                        </Sensortype>
#sensor_scaling#
#sensor_units_template#
                    </Sensor>
                    <SyncSignal address="1">
                        <OutputRate>#sample_rate#</OutputRate>
#filter_template#
                        <Output>
                            <RtAvailable>#rt_enable#</RtAvailable>
                        </Output>
                    </SyncSignal>
#signal2#
                    <OutputUnit>
#units_template#
                    </OutputUnit>
                    <ZeroOffset>
                        <Inhibit>0</Inhibit>
                        <Offset>#calibration_rezero_intercept#</Offset>
                        <Target>0.000000</Target>
                    </ZeroOffset>
                </AnalogInChannel>
            </Mode>
        </Connector#channel_num#>
    </Amplifier>
]],
    Resistor = [[
<?xml version="1.0" encoding="UTF-8"?>
    <Amplifier>
        <Connector#channel_num#>
            <Mode choice="AnalogInChannel">
                <AnalogInChannel>
                    <Sensor>
                        <Sensortype choice="Resistor">
                            <Resistor>
                                <MaxElectrical>5000</MaxElectrical>
                                <Wiring>4</Wiring>
                            </Resistor>
                        </Sensortype>
#sensor_scaling#
#sensor_units_template#
                    </Sensor>
                    <SyncSignal address="1">
                        <OutputRate>#sample_rate#</OutputRate>
#filter_template#
                        <Output>
                            <RtAvailable>#rt_enable#</RtAvailable>
                        </Output>
                    </SyncSignal>
#signal2#
                    <OutputUnit>
#units_template#
                    </OutputUnit>
                    <ZeroOffset>
                        <Inhibit>0</Inhibit>
                        <Offset>#calibration_rezero_intercept#</Offset>
                        <Target>0.000000</Target>
                    </ZeroOffset>
                </AnalogInChannel>
            </Mode>
        </Connector#channel_num#>
    </Amplifier>
]],
    Voltage = [[
<?xml version="1.0" encoding="UTF-8"?>
    <Amplifier>
        <Connector#channel_num#>
            <Mode choice="AnalogInChannel">
                <AnalogInChannel>
                    <Sensor>
                        <Sensortype choice="Voltage">
                            <Voltage>
                                <MaxElectrical>#sensor_input_max#</MaxElectrical>
                            </Voltage>
                        </Sensortype>
#sensor_scaling#
#transducer_power#
#sensor_units_template#
                    </Sensor>
                    <SyncSignal address="1">
                        <OutputRate>#sample_rate#</OutputRate>
#filter_template#
                        <Output>
                            <RtAvailable>#rt_enable#</RtAvailable>
                        </Output>
                    </SyncSignal>
#signal2#
                    <OutputUnit>
#units_template#
                    </OutputUnit>
                    <ZeroOffset>
                        <Inhibit>0</Inhibit>
                        <Offset>#calibration_rezero_intercept#</Offset>
                        <Target>0.000000</Target>
                    </ZeroOffset>
                </AnalogInChannel>
            </Mode>
        </Connector#channel_num#>
    </Amplifier>
]],
    IEPE = [[
<?xml version="1.0" encoding="UTF-8"?>
    <Amplifier>
        <Connector#channel_num#>
            <Mode choice="AnalogInChannel">
                <AnalogInChannel>
                    <Sensor>
                        <Sensortype choice="IEPE">
                            <IEPE>
                                <MaxElectrical>10</MaxElectrical>
                            </IEPE>
                        </Sensortype>
#sensor_scaling#
#sensor_units_template#
                    </Sensor>
                    <SyncSignal address="1">
                        <OutputRate>#sample_rate#</OutputRate>
#filter_template#
                        <Output>
                            <RtAvailable>#rt_enable#</RtAvailable>
                        </Output>
                    </SyncSignal>
#signal2#
                    <OutputUnit>
#units_template#
                    </OutputUnit>
                    <ZeroOffset>
                        <Inhibit>0</Inhibit>
                        <Offset>#calibration_rezero_intercept#</Offset>
                        <Target>0.000000</Target>
                    </ZeroOffset>
                </AnalogInChannel>
            </Mode>
        </Connector#channel_num#>
    </Amplifier>
]],
    Current = [[
<?xml version="1.0" encoding="UTF-8"?>
    <Amplifier>
        <Connector#channel_num#>
            <Mode choice="AnalogInChannel">
                <AnalogInChannel>
                    <Sensor>
                        <Sensortype choice="Current">
                            <Current>
                                <MaxElectrical>20.0</MaxElectrical>
                            </Current>
                        </Sensortype>
#sensor_scaling#
#transducer_power#
#sensor_units_template#
                    </Sensor>
                    <SyncSignal address="1">
                        <OutputRate>#sample_rate#</OutputRate>
#filter_template#
                        <Output>
                            <RtAvailable>#rt_enable#</RtAvailable>
                        </Output>
                    </SyncSignal>
#signal2#
                    <OutputUnit>
#units_template#
                    </OutputUnit>
                    <ZeroOffset>
                        <Inhibit>0</Inhibit>
                        <Offset>#calibration_rezero_intercept#</Offset>
                        <Target>0.000000</Target>
                    </ZeroOffset>
                </AnalogInChannel>
            </Mode>
        </Connector#channel_num#>
    </Amplifier>
]],
    Thermocouple = [[
<?xml version="1.0" encoding="UTF-8"?>
    <Amplifier>
        <Connector#channel_num#>
            <Mode choice="AnalogInChannel">
                <AnalogInChannel>
                    <Sensor>
                        <Sensortype choice="Thermocouple">
                            <Thermocouple>
                                <ThermocoupleType>#_thermocouple_type#</ThermocoupleType>
                            </Thermocouple>
                        </Sensortype>
                        <Scaling>
                            <Scalingtype choice="Internal">
                                <Internal/>
                            </Scalingtype>
                        </Scaling>
#sensor_units_template#
                    </Sensor>
                    <SyncSignal address="1">
                        <OutputRate>#sample_rate#</OutputRate>
#filter_template#
                        <Output>
                            <RtAvailable>#rt_enable#</RtAvailable>
                        </Output>
                    </SyncSignal>
#signal2#
                    <OutputUnit>
#units_template#
                    </OutputUnit>
                    <ZeroOffset>
                        <Inhibit>0</Inhibit>
                        <Offset>#calibration_rezero_intercept#</Offset>
                        <Target>0.000000</Target>
                    </ZeroOffset>
                </AnalogInChannel>
            </Mode>
        </Connector#channel_num#>
    </Amplifier>
]],
    LVDT = [[
<?xml version="1.0" encoding="UTF-8"?>
    <Amplifier>
        <Connector#channel_num#>
            <Mode choice="AnalogInChannel">
                <AnalogInChannel>
                    <Sensor>
                        <Sensortype choice="LVDT">
                            <LVDT>
                                <MaxElectrical>3000</MaxElectrical>
                            </LVDT>
                        </Sensortype>
#sensor_scaling#
#sensor_units_template#
                    </Sensor>
                    <SyncSignal address="1">
                        <OutputRate>#sample_rate#</OutputRate>
#filter_template#
                        <Output>
                            <RtAvailable>#rt_enable#</RtAvailable>
                        </Output>
                    </SyncSignal>
#signal2#
                    <OutputUnit>
#units_template#
                    </OutputUnit>
                    <ZeroOffset>
                        <Inhibit>0</Inhibit>
                        <Offset>#calibration_rezero_intercept#</Offset>
                        <Target>0.000000</Target>
                    </ZeroOffset>
                </AnalogInChannel>
            </Mode>
        </Connector#channel_num#>
    </Amplifier>
]],
    IndBridge = [[
<?xml version="1.0" encoding="UTF-8"?>
    <Amplifier>
        <Connector#channel_num#>
            <Mode choice="AnalogInChannel">
                <AnalogInChannel>
                    <Sensor>
                        <Sensortype choice="IndBridge">
                            <IndBridge>
                                <MaxElectrical>#sensor_input_max#</MaxElectrical>
                                <BridgeType>#_bridge_type#</BridgeType>
                                <Wiring>#_bridge_wiring#</Wiring>
                                <ExciteAmpl>#excitation_range#</ExciteAmpl>
                                <ExciteFreq>#excitation_mode#</ExciteFreq>
                            </IndBridge>
                        </Sensortype>
#sensor_scaling#
#sensor_units_template#
                    </Sensor>
                    <SyncSignal address="1">
                        <OutputRate>#sample_rate#</OutputRate>
#filter_template#
                        <Output>
                            <RtAvailable>#rt_enable#</RtAvailable>
                        </Output>
                    </SyncSignal>
#signal2#
                    <OutputUnit>
#units_template#
                    </OutputUnit>
                    <ZeroOffset>
                        <Inhibit>0</Inhibit>
                        <Offset>#calibration_rezero_intercept#</Offset>
                        <Target>0.000000</Target>
                    </ZeroOffset>
                </AnalogInChannel>
            </Mode>
        </Connector#channel_num#>
    </Amplifier>
]],
    Counter = [[
<?xml version="1.0" encoding="UTF-8"?>
    <Amplifier>
        <Connector#channel_num#>
            <Mode choice="AnalogInChannel">
                <AnalogInChannel>
                    <Sensor>
                        <Sensortype choice="Counter">
                            <Counter>
#optional_max_electrical#
                                <IndexEnbl>#_indexenbl#</IndexEnbl>
                                <IndexDivisor>#_indexdivisor#</IndexDivisor>
                                <S2Enbl>#_s2enbl#</S2Enbl>
                                <F4Enbl>#_f4enbl#</F4Enbl>
#optional_ppr#
#special_parameters#
                            </Counter>
                        </Sensortype>
#sensor_scaling#
#transducer_power#
#sensor_units_template#
                    </Sensor>
                    <SyncSignal address="1">
                        <OutputRate>#sample_rate#</OutputRate>
#filter_template#
                        <Output>
                            <RtAvailable>#rt_enable#</RtAvailable>
                        </Output>
                    </SyncSignal>
#connection_source#
#signal2#
                    <OutputUnit>
#units_template#
                    </OutputUnit>
                    <ZeroOffset>
                        <Inhibit>0</Inhibit>
                        <Offset>#calibration_rezero_intercept#</Offset>
                        <Target>0.000000</Target>
                    </ZeroOffset>
                </AnalogInChannel>
            </Mode>
        </Connector#channel_num#>
    </Amplifier>
]],
    SSI = [[
<?xml version="1.0" encoding="UTF-8"?>
    <Amplifier>
        <Connector#channel_num#>
            <Mode choice="AnalogInChannel">
                <AnalogInChannel>
                    <Sensor>
                        <Sensortype choice="SSI">
                            <SSI>
                                <SsiClockFreq>#_ssi_clock#</SsiClockFreq>
                                <SsiResolution>#_ssi_resolution#</SsiResolution>
                                <SsiCoding>#_ssi_coding#</SsiCoding>
                            </SSI>
                        </Sensortype>
#sensor_scaling#
#transducer_power#
#sensor_units_template#
                    </Sensor>
                    <SyncSignal address="1">
                        <OutputRate>#sample_rate#</OutputRate>
#filter_template#
                        <Output>
                            <RtAvailable>#rt_enable#</RtAvailable>
                        </Output>
                    </SyncSignal>
#signal2#
                    <OutputUnit>
#units_template#
                    </OutputUnit>
                    <ZeroOffset>
                        <Inhibit>0</Inhibit>
                        <Offset>#calibration_rezero_intercept#</Offset>
                        <Target>0.000000</Target>
                    </ZeroOffset>
                </AnalogInChannel>
            </Mode>
        </Connector#channel_num#>
    </Amplifier>
]],
    Frequency = [[
<?xml version="1.0" encoding="UTF-8"?>
    <Amplifier>
        <Connector#channel_num#>
            <Mode choice="AnalogInChannel">
                <AnalogInChannel>
                    <Sensor>
                        <Sensortype choice="Frequency">
                            <Frequency>
                                <MaxElectrical>#sensor_input_max#</MaxElectrical>
                                <S2Enbl>#_s2enbl#</S2Enbl>
                                <F4Enbl>#_f4enbl#</F4Enbl>
#special_parameters#
                            </Frequency>
                        </Sensortype>
#sensor_scaling#
#transducer_power#
#sensor_units_template#
                    </Sensor>
                    <SyncSignal address="1">
                        <OutputRate>#sample_rate#</OutputRate>
#filter_template#
                        <Output>
                            <RtAvailable>#rt_enable#</RtAvailable>
                        </Output>
                    </SyncSignal>
#connection_source#
#signal2#
                    <OutputUnit>
#units_template#
                    </OutputUnit>
                    <ZeroOffset>
                        <Inhibit>0</Inhibit>
                        <Offset>#calibration_rezero_intercept#</Offset>
                        <Target>0.000000</Target>
                    </ZeroOffset>
                </AnalogInChannel>
            </Mode>
        </Connector#channel_num#>
    </Amplifier>
]],
    -- The templates below are MX460 specific
    IndFrequency = [[
<?xml version="1.0" encoding="UTF-8"?>
    <Amplifier>
        <Connector#channel_num#>
            <Mode choice="AnalogInChannel">
                <AnalogInChannel>
                    <Sensor>
                        <Sensortype choice="IndFrequency">
                            <IndFrequency>
                                <MaxElectrical>#sensor_input_max#</MaxElectrical>
#special_parameters#
                            </IndFrequency>
                        </Sensortype>
#sensor_scaling#
#transducer_power#
#sensor_units_template#
                    </Sensor>
                    <SyncSignal address="1">
                        <OutputRate>#sample_rate#</OutputRate>
#filter_template#
                        <Output>
                            <RtAvailable>#rt_enable#</RtAvailable>
                        </Output>
                    </SyncSignal>
                    <ConnectorSource>#source_connection#</ConnectorSource>
#signal2#
                    <OutputUnit>
#units_template#
                    </OutputUnit>
                    <ZeroOffset>
                        <Inhibit>0</Inhibit>
                        <Offset>#calibration_rezero_intercept#</Offset>
                        <Target>0.000000</Target>
                    </ZeroOffset>
                </AnalogInChannel>
            </Mode>
        </Connector#channel_num#>
    </Amplifier>
]],
    IndCounter = [[
<?xml version="1.0" encoding="UTF-8"?>
    <Amplifier>
        <Connector#channel_num#>
            <Mode choice="AnalogInChannel">
                <AnalogInChannel>
                    <Sensor>
                        <Sensortype choice="IndCounter">
                            <IndCounter>
                                <MaxElectrical>2147484</MaxElectrical>
#special_parameters#
                            </IndCounter>
                        </Sensortype>
#sensor_scaling#
#transducer_power#
#sensor_units_template#
                    </Sensor>
                    <SyncSignal address="1">
                        <OutputRate>#sample_rate#</OutputRate>
                        <Output>
                            <RtAvailable>#rt_enable#</RtAvailable>
                        </Output>
                    </SyncSignal>
                    <ConnectorSource>#source_connection#</ConnectorSource>
#signal2#
                    <OutputUnit>
#units_template#
                    </OutputUnit>
                    <ZeroOffset>
                        <Inhibit>0</Inhibit>
                        <Offset>#calibration_rezero_intercept#</Offset>
                        <Target>0.000000</Target>
                    </ZeroOffset>
                </AnalogInChannel>
            </Mode>
        </Connector#channel_num#>
    </Amplifier>
]],
    Period = [[
<?xml version="1.0" encoding="UTF-8"?>
    <Amplifier>
        <Connector#channel_num#>
            <Mode choice="AnalogInChannel">
                <AnalogInChannel>
                    <Sensor>
                        <Sensortype choice="Period">
                            <Period>
                                <MaxElectrical>5000</MaxElectrical>
#special_parameters#
                            </Period>
                        </Sensortype>
#sensor_scaling#
#transducer_power#
#sensor_units_template#
                    </Sensor>
                    <SyncSignal address="1">
                        <OutputRate>#sample_rate#</OutputRate>
#filter_template#
                        <Output>
                            <RtAvailable>#rt_enable#</RtAvailable>
                        </Output>
                    </SyncSignal>
                    <ConnectorSource>#source_connection#</ConnectorSource>
#signal2#
                    <OutputUnit>
#units_template#
                    </OutputUnit>
                    <ZeroOffset>
                        <Inhibit>0</Inhibit>
                        <Offset>#calibration_rezero_intercept#</Offset>
                        <Target>0.000000</Target>
                    </ZeroOffset>
                </AnalogInChannel>
            </Mode>
        </Connector#channel_num#>
    </Amplifier>
]],
    PulseWidth = [[
<?xml version="1.0" encoding="UTF-8"?>
    <Amplifier>
        <Connector#channel_num#>
            <Mode choice="AnalogInChannel">
                <AnalogInChannel>
                    <Sensor>
                        <Sensortype choice="PulseWidth">
                            <PulseWidth>
                                <MaxElectrical>5000</MaxElectrical>
#special_parameters#
                            </PulseWidth>
                        </Sensortype>
#sensor_scaling#
#transducer_power#
#sensor_units_template#
                    </Sensor>
                    <SyncSignal address="1">
                        <OutputRate>#sample_rate#</OutputRate>
#filter_template#
                        <Output>
                            <RtAvailable>#rt_enable#</RtAvailable>
                        </Output>
                    </SyncSignal>
                    <ConnectorSource>#source_connection#</ConnectorSource>
#signal2#
                    <OutputUnit>
#units_template#
                    </OutputUnit>
                    <ZeroOffset>
                        <Inhibit>0</Inhibit>
                        <Offset>#calibration_rezero_intercept#</Offset>
                        <Target>0.000000</Target>
                    </ZeroOffset>
                </AnalogInChannel>
            </Mode>
        </Connector#channel_num#>
    </Amplifier>
]],
    Pwm = [[
<?xml version="1.0" encoding="UTF-8"?>
    <Amplifier>
        <Connector#channel_num#>
            <Mode choice="AnalogInChannel">
                <AnalogInChannel>
                    <Sensor>
                        <Sensortype choice="Pwm">
                            <Pwm>
                                <MaxElectrical>100</MaxElectrical>
#special_parameters#
                            </Pwm>
                        </Sensortype>
#sensor_scaling#
#transducer_power#
#sensor_units_template#
                    </Sensor>
                    <SyncSignal address="1">
                        <OutputRate>#sample_rate#</OutputRate>
#filter_template#
                        <Output>
                            <RtAvailable>#rt_enable#</RtAvailable>
                        </Output>
                    </SyncSignal>
                    <ConnectorSource>#source_connection#</ConnectorSource>
#signal2#
                    <OutputUnit>
#units_template#
                    </OutputUnit>
                    <ZeroOffset>
                        <Inhibit>0</Inhibit>
                        <Offset>#calibration_rezero_intercept#</Offset>
                        <Target>0.000000</Target>
                    </ZeroOffset>
                </AnalogInChannel>
            </Mode>
        </Connector#channel_num#>
    </Amplifier>
]],
    CrankWheelFrequency = [[
<?xml version="1.0" encoding="UTF-8"?>
    <Amplifier>
        <Connector#channel_num#>
            <Mode choice="AnalogInChannel">
                <AnalogInChannel>
                    <Sensor>
                        <Sensortype choice="CrankWheelFrequency">
                            <CrankWheelFrequency>
                                <MaxElectrical>#sensor_input_max#</MaxElectrical>
                                <PulsesPerRound>#sensor_input_pulses_per_round#</PulsesPerRound>
                                <CountingEdge>#sensor_input_pulse_edge#</CountingEdge>
                                <GapPosition>#sensor_input_gap_position#</GapPosition>
#special_parameters#
                            </CrankWheelFrequency>
                        </Sensortype>
#sensor_scaling#
#transducer_power#
#sensor_units_template#
                    </Sensor>
                    <SyncSignal address="1">
                        <OutputRate>#sample_rate#</OutputRate>
#filter_template#
                        <Output>
                            <RtAvailable>#rt_enable#</RtAvailable>
                        </Output>
                    </SyncSignal>
                    <ConnectorSource>#source_connection#</ConnectorSource>
#signal2#
                    <OutputUnit>
#units_template#
                    </OutputUnit>
                    <ZeroOffset>
                        <Inhibit>0</Inhibit>
                        <Offset>#calibration_rezero_intercept#</Offset>
                        <Target>0.000000</Target>
                    </ZeroOffset>
                </AnalogInChannel>
            </Mode>
        </Connector#channel_num#>
    </Amplifier>
]],
    CrankWheelCounter = [[
<?xml version="1.0" encoding="UTF-8"?>
    <Amplifier>
        <Connector#channel_num#>
            <Mode choice="AnalogInChannel">
                <AnalogInChannel>
                    <Sensor>
                        <Sensortype choice="CrankWheelCounter">
                            <CrankWheelCounter>
                                <MaxElectrical>2147484</MaxElectrical>
                                <PulsesPerRound>#sensor_input_pulses_per_round#</PulsesPerRound>
                                <CountingEdge>#sensor_input_pulse_edge#</CountingEdge>
                                <GapPosition>#sensor_input_gap_position#</GapPosition>
                                <IndexPosition>#sensor_input_zero_position#</IndexPosition>
                                <IndexEdge>#sensor_input_zero_edge#</IndexEdge>
                                <IndexDivisor>1</IndexDivisor>
#special_parameters#
                            </CrankWheelCounter>
                        </Sensortype>
#sensor_scaling#
#transducer_power#
#sensor_units_template#
                    </Sensor>
                    <SyncSignal address="1">
                        <OutputRate>#sample_rate#</OutputRate>
                        <Output>
                            <RtAvailable>#rt_enable#</RtAvailable>
                        </Output>
                    </SyncSignal>
                    <ConnectorSource>#source_connection#</ConnectorSource>
#signal2#
                    <OutputUnit>
#units_template#
                    </OutputUnit>
                    <ZeroOffset>
                        <Inhibit>0</Inhibit>
                        <Offset>#calibration_rezero_intercept#</Offset>
                        <Target>0.000000</Target>
                    </ZeroOffset>
                </AnalogInChannel>
            </Mode>
        </Connector#channel_num#>
    </Amplifier>
]],
    IndCrankWheelFrequency = [[
<?xml version="1.0" encoding="UTF-8"?>
    <Amplifier>
        <Connector#channel_num#>
            <Mode choice="AnalogInChannel">
                <AnalogInChannel>
                    <Sensor>
                        <Sensortype choice="IndCrankWheelFrequency">
                            <IndCrankWheelFrequency>
                                <MaxElectrical>#sensor_input_max#</MaxElectrical>
                                <PulsesPerRound>#sensor_input_pulses_per_round#</PulsesPerRound>
                                <CountingEdge>#sensor_input_pulse_edge#</CountingEdge>
                                <GapPosition>#sensor_input_gap_position#</GapPosition>
#special_parameters#
                            </IndCrankWheelFrequency>
                        </Sensortype>
#sensor_scaling#
#transducer_power#
#sensor_units_template#
                    </Sensor>
                    <SyncSignal address="1">
                        <OutputRate>#sample_rate#</OutputRate>
#filter_template#
                        <Output>
                            <RtAvailable>#rt_enable#</RtAvailable>
                        </Output>
                    </SyncSignal>
                    <ConnectorSource>#source_connection#</ConnectorSource>
#signal2#
                    <OutputUnit>
#units_template#
                    </OutputUnit>
                    <ZeroOffset>
                        <Inhibit>0</Inhibit>
                        <Offset>#calibration_rezero_intercept#</Offset>
                        <Target>0.000000</Target>
                    </ZeroOffset>
                </AnalogInChannel>
            </Mode>
        </Connector#channel_num#>
    </Amplifier>
]],
    IndCrankWheelCounter = [[
<?xml version="1.0" encoding="UTF-8"?>
    <Amplifier>
        <Connector#channel_num#>
            <Mode choice="AnalogInChannel">
                <AnalogInChannel>
                    <Sensor>
                        <Sensortype choice="IndCrankWheelCounter">
                            <IndCrankWheelCounter>
                                <MaxElectrical>2147484</MaxElectrical>
                                <PulsesPerRound>#sensor_input_pulses_per_round#</PulsesPerRound>
                                <CountingEdge>#sensor_input_pulse_edge#</CountingEdge>
                                <GapPosition>#sensor_input_gap_position#</GapPosition>
                                <IndexPosition>#sensor_input_zero_position#</IndexPosition>
                                <IndexEdge>#sensor_input_zero_edge#</IndexEdge>
                                <IndexDivisor>1</IndexDivisor>
#special_parameters#
                            </IndCrankWheelCounter>
                        </Sensortype>
#sensor_scaling#
#transducer_power#
#sensor_units_template#
                    </Sensor>
                    <SyncSignal address="1">
                        <OutputRate>#sample_rate#</OutputRate>
                        <Output>
                            <RtAvailable>#rt_enable#</RtAvailable>
                        </Output>
                    </SyncSignal>
                    <ConnectorSource>#source_connection#</ConnectorSource>
#signal2#
                    <OutputUnit>
#units_template#
                    </OutputUnit>
                    <ZeroOffset>
                        <Inhibit>0</Inhibit>
                        <Offset>#calibration_rezero_intercept#</Offset>
                        <Target>0.000000</Target>
                    </ZeroOffset>
                </AnalogInChannel>
            </Mode>
        </Connector#channel_num#>
    </Amplifier>
]],
}

local transducer_power_template = [[
                        <Supply>#transducer_power#</Supply>
]]

local scaling_templates = {
    defined_two_point = [[
                        <Scaling>
                            <Scalingtype choice="LinTable">
                                <LinTable>
                                    <Count>2</Count>
                                    <Point address="1">
                                        <X>#calibration_sensor_scaling_electrical_1#</X>
                                        <Y>#calibration_sensor_scaling_physical_1#</Y>
                                    </Point>
                                    <Point address="2">
                                        <X>#calibration_sensor_scaling_electrical_2#</X>
                                        <Y>#calibration_sensor_scaling_physical_2#</Y>
                                    </Point>
                                </LinTable>
                            </Scalingtype>
                        </Scaling>
]],
    defined_zero_span = [[
                        <Scaling>
                            <Scalingtype choice="ZeroSpan">
                                <ZeroSpan>
                                    <ElecZero>#calibration_sensor_scaling_electrical_zero#</ElecZero>
                                    <ElecSpan>#calibration_sensor_scaling_electrical_span#</ElecSpan>
                                    <PhysSpan>#calibration_sensor_scaling_physical_span#</PhysSpan>
                                </ZeroSpan>
                            </Scalingtype>
                        </Scaling>
]],
    defined_slope_intercept = [[
                        <Scaling>
                            <Scalingtype choice="Gain">
                                <Gain>
                                    <m>#calibration_sensor_scaling_slope#</m>
                                    <b>#calibration_sensor_scaling_intercept#</b>
                                </Gain>
                            </Scalingtype>
                        </Scaling>
]],
    strain_gage = [[
                        <Scaling>
                            <Scalingtype choice="Straingage">
                                <Straingage>
                                    <kFactor>#calibration_sensor_scaling_gage_factor#</kFactor>
                                    <Bridgefactor>#calibration_sensor_scaling_bridge_factor#</Bridgefactor>
                                </Straingage>
                            </Scalingtype>
                        </Scaling>
]],
    internal = [[
                        <Scaling>
                            <Scalingtype choice="Internal">
                                <Internal/>
                            </Scalingtype>
                        </Scaling>
]],
}

local filter_templates = {
    Off = [[
                        <FilterType choice="Off">
                            <Off />
                        </FilterType>
]],
    Lowpass = [[
                        <FilterType choice="Lowpass">
                            <Lowpass>
                                <FiltCharact>#digital_filter_type#</FiltCharact>
                                <FiltFreq>#digital_filter_attenuation_frequency#</FiltFreq>
                            </Lowpass>
                        </FilterType>
]],
}

local units_templates = {
    UserDefined =
[[                        <UnitId>100000</UnitId>
                        <UserText>#calibration_units_conversion_text#</UserText>
                        <Factor>#calibration_units_conversion_slope#</Factor>
                        <Offset>#calibration_units_conversion_intercept#</Offset>
]],
    Predefined =
[[                        <UnitId>#calibration_units_conversion_units#</UnitId>
]],
}

local sensor_units_templates = {
    UserDefined =
[[                        <PhysUnit>100000</PhysUnit>
                        <PhysUnitUsr>#calibration_define_source_units_text#</PhysUnitUsr>
]],
    Predefined = 
[[                        <PhysUnit>#calibration_define_source_units_units#</PhysUnit>
]],
}

function Configure:_setup_sensor(xtemplate, chan_num, spec, cb)
    local special_parameters = spec.special_parameters
    local spec = flatten(spec)

    local template =
        gsub_literal(xtemplate, '#channel_num#', tostring(chan_num))

    -- Add in sensor scaling stuff if needed
    if template:find('#sensor_scaling#') then
        local scaling_template =
            scaling_templates[spec.calibration_sensor_scaling_mode]
        if not scaling_template then
            if cb then
                cb({
                    result = 'error',
                    msg = string.format("Can not find scaling template for %s",
                        spec.calibration_sensor_scaling_mode),
                })
            end
            return
        end
        template = gsub_literal(template, '#sensor_scaling#', scaling_template)
    end

    -- Add transducer power if needed.  NOTE:  Transducer power is only valid
    -- on some MX1601 channels!
    if template:find('#transducer_power#') then
        local t = ''
        if spec.transducer_power then
            t = transducer_power_template
        end
        template = gsub_literal(template, '#transducer_power#', t)
    end

    -- Add the ConnectorSource stuff if needed.  This is for the MX460
    if template:find('#connection_source#') then
        if not spec.source_connection then
            template = gsub_literal(template, '#connection_source#', '')
        else
            template = gsub_literal(template, '#connection_source#',
                '<ConnectorSource>#source_connection#</ConnectorSource>')
        end
    end

    -- Add pulses per round to Counter template.  This only happens with an
    -- MX460.
    if template:find('#optional_ppr#') then
        if not spec.sensor_input_pulses_per_round then
            template = gsub_literal(template, '#optional_ppr#', '')
        else
            template = gsub_literal(template, '#optional_ppr#',
                '<PulsesPerRound>#sensor_input_pulses_per_round#</PulsesPerRound>')
        end
    end

    -- Add MaxElectrical to "Counter" if needed
    if template:find("#optional_max_electrical#") then
        if not spec.sensor_input_max then
            template = gsub_literal(template, '#optional_max_electrical#', '')
        else
            template = gsub_literal(template, '#optional_max_electrical#',
                '<MaxElectrical>#sensor_input_max#</MaxElectrical>')
        end
    end

    -- Add special parameters (SpecialParameters) if needed (MX460)
    if not special_parameters then
        template = gsub_literal(template, '#special_parameters#', '')
    else
        local mapping = {
            zero_shift = {
                'IndexShift',
                function(v) return tostring(v) end
            },
            interpolation_enable = {
                'InterpolationEnbl',
                function(v) return v and '1' or '0' end
            },
            glitch_filter = {
                'GlitchFilter',
                function(v) return tostring(v) end
            },
            invert_polarity = {
                'Polarity',
                function(v) return v and '1' or '0' end
            },
            termination = {
                'Termination',
                function(v) return v and '1' or '0' end
            },
        }
        local param_sets = {
            pulse_frequency = {
                glitch_filter = true,
                invert_polarity = true,
                termination = true,
            },
            encoder_frequency = {
                glitch_filter = true,
                invert_polarity = true,
                termination = true,
            },
            pulse_counter = {
                zero_shift = true,
                interpolation_enable = true,
                glitch_filter = true,
                invert_polarity = true,
                termination = true,
            },
            encoder_counter = {
                zero_shift = true,
                interpolation_enable = true,
                glitch_filter = true,
                invert_polarity = true,
                termination = true,
            },
            inductive_pulse_frequency = {
                glitch_filter = true,
            },
            inductive_pulse_counter = {
                interpolation_enable = true,
                glitch_filter = true,
                invert_polarity = true,
            },
            pulse_period = {
                glitch_filter = true,
                invert_polarity = true,
                termination = true,
            },
            pulse_high_period = {
                glitch_filter = true,
                invert_polarity = true,
                termination = true,
            },
            pulse_duty_cycle = {
                glitch_filter = true,
                invert_polarity = true,
                termination = true,
            },
            crank_wheel_frequency = {
                glitch_filter = true,
            },
            crank_wheel_counter = {
                zero_shift = true,
                interpolation_enable = true,
                glitch_filter = true,
            },
            inductive_crank_wheel_frequency = {
                glitch_filter = true,
            },
            inductive_crank_wheel_counter = {
                zero_shift = true,
                interpolation_enable = true,
                glitch_filter = true,
            },
        }

        local pset = param_sets[spec.sensor_input_mode] or {}
        local str = "<SpecialParameters>\n"
        for k, v in pairs(special_parameters) do
            local m = mapping[k]
            if m and pset[k] then
                str = str .. string.format("\t<%s>%s</%s>\n",
                    m[1], m[2](v), m[1])
            end
        end
        str = str .. "</SpecialParameters>\n"
        template = gsub_literal(template, '#special_parameters#', str)
    end

    -- Add signal2 stuff
    if template:find('#signal2#') then
        if not spec.signal2_sample_rate then
            -- Disable signal2 on firewire
            template = gsub_literal(template, '#signal2#',
                self.signal2_templates.disabled)
        else
            -- Enable signal2 on firewire
            template = gsub_literal(template, '#signal2#',
                self.signal2_templates.enabled)
            template = gsub_literal(template, '#signal2_sample_rate#',
                spec.signal2_sample_rate)
            if not spec.signal2_digital_filter_type or
                spec.signal2_digital_filter_type == 'none' then
                template = gsub_literal(template, '#signal2_filter#',
                    filter_templates.Off)
            else
                template = gsub_literal(template, '#signal2_filter#',
                    filter_templates.Lowpass)
                local ftypes = {
                    bessel_iir = 1,
                    butterworth_iir = 2,
                    linear_phase_fir = 3,
                    butterworth_fir = 4,
                }
                template = gsub_literal(template,
                    '#digital_filter_type#',
                    ftypes[spec.signal2_digital_filter_type])
                template = gsub_literal(template,
                    '#digital_filter_attenuation_frequency#',
                    spec.signal2_digital_filter_attenuation_frequency)
            end
        end
    end

    -- Add in filter stuff if needed.  This includes sample rate
    if template:find('#filter_template#') then
        local filter_template =
            filter_templates[spec._digital_filter_choice]
        if not filter_template then
            if cb then
                cb({
                    result = 'error',
                    msg = string.format("Can not find filter template for %s",
                        spec._digital_filter_choice),
                })
            end
            return
        end
        template = gsub_literal(template, '#filter_template#', filter_template)
    end

    -- Add in unit conversion stuff
    if template:find('#units_template#') then
        local units_template =
            units_templates[spec.calibration_units_conversion_units] or
            units_templates['Predefined']
        if not units_template then
            if cb then
                cb({
                    result = 'error',
                    msg = string.format(
                        "Can not find units conversion template for %s",
                        spec.calibration_units_conversion_text),
                })
            end
            return
        end
        template = gsub_literal(template, '#units_template#', units_template)
    end

    -- Add in sensor units stuff
    if template:find('#sensor_units_template#') then
        local units_template =
            sensor_units_templates[spec.calibration_define_source_units_units] or
            sensor_units_templates['Predefined']
        if not units_template then
            if cb then
                cb({
                    result = 'error',
                    msg = string.format(
                        "Can not find sensor units template for %s",
                        spec.calibration_define_source_units_text),
                })
            end
            return
        end
        template = gsub_literal(template, '#sensor_units_template#',
            units_template)
    end

    -- Do all of the parameter substitution on the resulting XML template
    for k, v in pairs(spec) do
        if type(v) == 'string' or type(v) == 'number' then
            local value = v
            if self.parameter_translator[k] then
                value = self.parameter_translator[k](self, v) or v
            end
            template = gsub_literal(template, string.format('#%s#', k),  value)
        end
    end

    -- Sanity check for any untranslated parameters
    local untranslated = template:match('#([_a-zA-Z0-9]+)#')
    if untranslated then
        if cb then
            cb({
                result = 'error',
                msg = string.format("Untranslated MX XML parameter '%s'",
                    untranslated)
            })
        end
        return
    end

    -- Do encoding
    for k, v in pairs{['>'] = '&gt;', ['<'] = '&lt;', ['"'] = '&quot;'} do
        template = gsub_literal(template, k, v)
    end

    -- Finish it up so it can be sent in a transaction request
    local xml = string.format('<?xml version="1.0" encoding="UTF-8"?>\n<SrvParams><View>5</View><XmlSection>%s</XmlSection></SrvParams>', template)

    self:_send_transaction_request(C.TR_DUNNO_0, C.TR_PLAIN_XML,
        function(obj, ecode, data, all_errors)
            local results = generate_mx_results(
                string.format('Unable to set parameters for channel %d',
                    chan_num), xml, ecode, all_errors)
            if cb then cb(results) end
        end, xml)
end


-- Function used to configure channels
function Configure:setup_channel(chan_num, spec, cb)
    local teds_edited = teds_utils.is_teds_edited(spec)
    local spec, err = self:_convert_spec_to_internal(spec)
    if not spec then
        local res = {
            result = 'error',
            msg = err or 'Unable to convert channel setup to internal format',
        }
        if cb then cb(res) end
        return res
    end
    spec.teds_edited = teds_edited

    if not spec.sensor_input or not spec.sensor_input.mode then
        return self:_setup_channel(chan_num, spec, cb)
    end

    local sensor_input_mode =
        self.parameter_translator:sensor_input_mode(spec.sensor_input.mode)

    -- If there is a sensor template, setup the parameters for it as a group
    local template = self.sensor_templates[sensor_input_mode]
    if template then
        self:_setup_sensor(template, chan_num, spec, cb)
        return {
            result = 'success',
            msg = 'Channel configuration in progress'
        }
    else
        return {
            result = 'error',
            msg = "Don't have sensor type template for " .. sensor_input_mode
        }
    end
end

-- Function use to complete channel configuration
function Configure:_setup_channel(chan_num, spec, cb)

    spec = flatten(spec)

    if not self.device.json2xpath then
        local r = {result = 'success', msg = 'Channel configuration complete'}
        if cb then cb(r) end
        return r
    end

    local j2x = flatten(self.device.json2xpath)
    local ignore_list = flatten(self.device.ignore_list)
    local value
    local response_list = {}

    local sensor_input_mode = self.parameter_translator:sensor_input_mode(
        spec.sensor_input_mode) or spec.sensor_input_mode or ''

    local results = {result = 'success', msg = ''}

    local xml = ''
    for p, v in pairs(spec) do
        local xpath = j2x[p]
        if xpath and not ignore_list[p] then
            response_list[#response_list + 1] = p
        end
    end

    -- Parameter must be setup in the proper order!
    local setup_order = {
        'sensor_input_mode',
        'sensor_input_type',
        'calibration_define_source_units_units',
        'calibration_sensor_scaling_mode',
        'sample_rate',
        'digital_filter_attenumation_frequency',
        '_digital_filter_choice',
    }
    local setup_spec = {}
    for i, k in ipairs(setup_order) do
        if spec[k] then
            setup_spec[#setup_spec + 1] = {p = k, v = spec[k]}
            spec[k] = nil
        end
    end
    -- Remainder of parameters, order doesn't matter
    for k, v in pairs(spec) do
        setup_spec[#setup_spec + 1] = {p = k, v = v}
    end

    -- Now send off all of the stuff to the MX device
    for _, kv in ipairs(setup_spec) do
        local p = kv.p
        local v = kv.v
        local xpath = j2x[p]
        if xpath and not ignore_list[p] then
            xpath = gsub_literal(xpath, '#channel_num#', tostring(chan_num))
            xpath = gsub_literal(xpath, '#sensor_input_mode#', sensor_input_mode)

            local value = v
            if self.parameter_translator[p] then
                value = self.parameter_translator[p](self, v) or v
            end
            local attr = xpath:match('/@(.-)$')
            if attr then
                xpath = gsub_literal(xpath, string.format('/@%s$', attr), '')
                xpath = string.format('%s[@%s="%s"]', xpath, attr, value)
                value = ''
            end
            local xml = xml_declaration ..
                '<SrvParams>' ..
                    '<View>5</View>' ..
                    '<Path>' .. xpath .. '</Path>' ..
                    '<XmlSection>' .. value .. '</XmlSection>' ..
                '</SrvParams>'
            self:_send_transaction_request(C.TR_DUNNO_0,
                C.TR_SET_CHANNEL_PARAMETER,
                function(obj, ecode, data, all_errors)
                    if ecode ~= C.ERR_SUCCESS then
                        results = generate_mx_results(
                            string.format('Unable to set %s for channel %d to %s',
                                p, chan_num, tostring(v)),
                            xml, ecode, all_errors)
                    end
                    table.remove(response_list, 1)
                    if #response_list == 0 then
                        if results.result == 'success' then
                            results.msg = 'All parameters successfully set'
                        end
                        if cb then cb(results) end
                    end
                end, xml)
        end
    end

    return {result = 'success', msg = 'Channel configuration in progress'}
end


-- Function used to setup all of the channels on the device to their default
-- settings
function Configure:_set_channel_defaults(chan_num)
    if chan_num and type(chan_num) == 'number' then
        self.chan_num = chan_num
    end
    if self.chan_num < self.num_channels then
        local dev_xml = self.device:get_default_xml(self.chan_num):gsub('<',
            '&lt;'):gsub('>', '&gt;'):gsub('"', '&quot;')

        local xml =
[[<?xml version="1.0" encoding="UTF-8"?>
    <SrvParams>
        <View>5</View>
            <XmlSection>#dev_xml#
        </XmlSection>
    </SrvParams>]]
        xml = gsub_literal(xml, '#dev_xml#', dev_xml)
        self.chan_num = self.chan_num + 1
        local cb
        if self.chan_num >= self.num_channels then
            -- When the response from the final channel is received, get the
            -- view information one more time just in case it changed after
            -- setting the channels to their default configurations.
            cb = self._get_views
            self._get_view_complete_cb = self._make_channel_info
        else
            cb = function(self, ecode, data, all_errors)
                    self:_set_channel_defaults(self.chan_num)
                end
        end
        self:_send_transaction_request(C.TR_DUNNO_0, C.TR_PLAIN_XML, cb, xml)
        return
    else
        self._get_view_complete_cb = self._make_channel_info
        self:_get_views()
    end
end

-- Function used to handle the response to the get views transaction request
function Configure:_handle_get_view_normal(ecode, buf, all_errors)
    -- Shove data into view info
    self.view[self.current_id] = lxp.lom.parse(ffi.string(buf))

    -- Get XML for next view
    local id, desc = self:_get_view_info()
    if id and desc then
        local xml = string.format(get_param_xml_fmt, id)
        self.current_id = desc
        self:_send_transaction_request(C.TR_DUNNO_0, C.TR_XML,
            self._handle_get_view, xml)
    elseif self._get_view_complete_cb then
        self:_get_view_complete_cb()
        self._get_view_complete_cb = nil
    end
end

local drd_map = {
    [0] = 'hbm_legacy',
    [2] = 'decimal',
    [3] = 'binary',
}

local set_alias_xml_template =
[[<?xml version="1.0"?>
<SrvParams>
    <View>3</View>
    <Path>System/ModuleName</Path>
    <XmlSection>#alias#</XmlSection>
</SrvParams>]]

-- Function used to set the module name (alias) for the device
function Configure:set_alias(alias)
    local xml = gsub_literal(set_alias_xml_template, '#alias#', alias)
    xml = gsub_literal(xml, '"', '&quot;')
    self:_send_transaction_request(C.TR_DUNNO_0, C.TR_SET_CHANNEL_PARAMETER,
        nil, xml)
end

local disable_rt_template =
[[<?xml version="1.0"?>
<SrvParams>
    <View>5</View>
    <Path>Amplifier/Connector#cnum#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/SyncSignal[@address="#signal#"]/Output/RtAvailable</Path>
    <XmlSection>0</XmlSection>
</SrvParams>]]

-- Function used to disable signals on the realtime (firewire) stream.  Note,
-- this does NOT work for CAN signal channels from MX471 devices.
-- FIXME:  Maybe make this work with MX471 CAN signal channels.
function Configure:disable_realtime_signals(cnum)
    local xml = gsub_literal(disable_rt_template, '#cnum#', cnum)
    for signal = 1, 2 do
        local x = gsub_literal(xml, '#signal#', tostring(signal))
        x = gsub_literal(x, '"', '&quot;')
        self:_send_transaction_request(C.TR_DUNNO_0,
            C.TR_SET_CHANNEL_PARAMETER, nil, x)
    end
end


local set_drd_xml_template = 
[[<?xml version="1.0"?>
<SrvParams>
    <View>3</View>
    <Path>System/DataRateDomain/Request</Path>
    <XmlSection>#drd#</XmlSection>
</SrvParams>]]

-- Function used to set the data rate domain of the device
function Configure:set_drd(state)
    if state == self.current_drd then
        local result = {
            result = 'success',
            msg = string.format("DRD is already set to %s", state),
        }
        return result
    end
    local drd_code
    for drd, v in pairs(drd_map) do
        if state == v then
            drd_code = drd
            break
        end
    end
    if not drd_code then
        local result = {
            result = 'error',
            msg = string.format("Unknown DRD %s", state),
        }
        self.logger:log(self.logger.ERR, result.msg)
        return result
    end
    local timer = ev.Timer.new(function(l, t, e)
                    local cmd = string.format(
                        "DROPBEAR_PASSWORD=hbmquantumx " ..
                        "/usr/bin/dbclient -y hbmuser@%s " ..
                        "'sync;sync;/sbin/reboot'", self.host)
                    self.logger.log = function() end
                    os.execute(cmd)
                    os.exit(1)
                end, 5, 0)

    self:set_led("flashing_red")
    local xml = gsub_literal(set_drd_xml_template, '#drd#', drd_code)
    xml = gsub_literal(xml, '"', '&quot;')
    self:_send_transaction_request(C.TR_DUNNO_0, C.TR_SET_CHANNEL_PARAMETER,
        function(obj, ecode, data, all_errors) timer:start(obj.loop) end,
        xml)

    return {
        result = 'success',
        msg = string.format("DRD change from %s to %s in progress",
            self.current_drd, state),
    }
end

-- Function used to get the firmware version from the firmware image file on
-- the disk
local function get_expected_fwver()
    local f = io.popen("ls -lt /usr/share/quantum-firmware/*.hbg 2>/dev/null")
    if f then
        for line in f:lines() do
            local e = line:match('firmware/(%d+-%d+-%d+-%d+)')
            if e then
                f:close()
                return e:gsub('-', '.')
            end
        end
        f:close()
    end
end

-- Function used to set the synchronization mode of the device
function Configure:_set_synchronization_mode(no_ptp)
    local xml = 
[[<?xml version="1.0"?>
<SrvParams>
    <View>3</View>
    <Path>System/TimeSource[@choice="Ptp"]/Ptp/Transport[@choice="UDPIPV6"]</Path>
    <XmlSection><UDPIPV6 /></XmlSection>
</SrvParams>]]

    -- If PTP should not be enabled on this device, set it to auto
    if no_ptp then
        xml = 
[[<?xml version="1.0"?>
<SrvParams>
    <View>3</View>
    <Path>System/TimeSource[@choice="Auto"]</Path>
    <XmlSection><Auto /></XmlSection>
</SrvParams>]]
    end

    xml = gsub_literal(xml, '"', '&quot;')
    self:_send_transaction_request(C.TR_DUNNO_0, C.TR_SET_CHANNEL_PARAMETER,
        function(obj, ecode, data, all_errors)
            -- Next stage of initialization load specific personality module and
            -- set all channels to their default configurations.
            local v = self.view['System parameters']
            self.devinfo.name = xpath.selectNodes(v,
                '//ModuleType/text()')[1]
            self.devinfo.fwver = xpath.selectNodes(v,
                '//FirmwareVersion/text()')[1]
            self.devinfo.serial_number = xpath.selectNodes(v,
                '//SerialNumber/text()')[1]
            self.devinfo.current_drd = xpath.selectNodes(v,
                '//DataRateDomain/CurrentState/text()')[1]
            self.devinfo.ip_addr = self.host
            if self.devinfo.name and #self.devinfo.name and
                tonumber(self.devinfo.fwver:match('^(%d+)%.')) > 2 and
                not self.devinfo.name:find('%d+B') then
                    self.logger:log(self.logger.DEBUG, string.format(
                        'Module %s at %s did not id itself as "B" model device',
                        self.devinfo.name, self.host))
                    self.devinfo.name = self.devinfo.name .. 'B'
            end
            self.devinfo.name = self.devinfo.name:lower()
            -- MX840 identifies itself with a quirky ModuleType string of
            -- "MX840A"; this results in a devinfo.name string of "mx840ab";
            -- concluded after discussions with DA that this name should be
            -- coerced to "mx840b" (see Redmine #4467).
            if self.devinfo.name == 'mx840ab' then
                self.devinfo.name = 'mx840b'
            end
            self.devinfo.expected_fwver =
                get_expected_fwver() or 'FW image not found'

            -- In case the 878 channels are OFF
            if self.devinfo.name:find('mx878') and
                (not self.num_channels or self.num_channels == 0) then
                self.num_channels = 8
            end

            if config.dump_mx_info then
                local fh = io.open(string.format('/tmp/mx-%s-%s-views-xml.txt',
                    self.devinfo.name, self.devinfo.serial_number), 'w')
                for _id, _v in pairs(self.raw_view) do
                    fh:write(_id .. '\n')
                    fh:write(_v .. '\n\n')
                end
                io.close(fh)
                self.raw_view = nil
            end
    
            local ok, device = pcall(require, 'mx.' .. self.devinfo.name)
            if ok then
                self.device = device

                -- Get any device specific translators
                underscore.extend(self.parameter_translator,
                    device.translator or {})
                device.translate = trans.translate_function()

                self:_set_channel_defaults(1)
                self._handle_get_view = self._handle_get_view_normal
            else
                self.logger:log(self.logger.ERR, string.format(
                    "Unable to find support module for %s : %s",
                    self.devinfo.name, device))
            end
        end, xml)
end

-- Function used to handle the response to the get views transaction request
function Configure:_handle_get_view_init(ecode, buf, all_errors)
    -- Shove data into view info
    local str = ffi.string(buf)
    if config.dump_mx_info then
        self.raw_view = self.raw_view or {}
        self.raw_view[self.current_id] = str
    end
    if str:find('MX878') then self.no_ptp = true end
    self.view[self.current_id] = lxp.lom.parse(str)
    
    -- Get XML for next view
    local id, desc = self:_get_view_info()
    if id and desc then
        local xml = string.format(get_param_xml_fmt, id)
        self.current_id = desc
        self.view[desc] = nil
        self:_send_transaction_request(C.TR_DUNNO_0, C.TR_XML,
            self._handle_get_view, xml)
    else
        -- Next stage of initialization is to set the device to use PTPv2/IPv6
        self:_set_synchronization_mode(self.no_ptp)
    end
end

Configure._handle_get_view = Configure._handle_get_view_init

-- Function used to retrieve the MX module type (name)
function Configure:get_module_type()
    return self.devinfo.name or 'unknown'
end

-- Function used to get the view information at initialization time
function Configure:_get_views(ecode, buf, all_errors)
    self.view = self.view or {}
    local id, desc = self:_get_view_info(true)
    if id and desc then
        self.current_id = desc
        self.view[desc] = nil
        local xml = string.format(get_param_xml_fmt, id)
        self:_send_transaction_request(C.TR_DUNNO_0, C.TR_XML,
            self._handle_get_view, xml)
    end
end

-- Function used to get one or more of the views
function Configure:get_views(views, cb)
    local vid_map = {
        ['Communication parameters'] = '1',
        ['System parameters'] = '3',
        ['Initial adjustment'] = '4',
        ['Analog input connector parameters'] = '5',
        ['Additional functions parameters'] = '6',
        ['Additional connectors parameters'] = '8',
    }
    for i, desc in pairs(views) do
        local id = vid_map[desc]
        if id then
            local xml = string.format(get_param_xml_fmt, id)
            self.view[desc] = nil
            self:_send_transaction_request(C.TR_DUNNO_0, C.TR_XML,
                function(s, ecode, buf, all_errors)
                    self.view[desc] = lxp.lom.parse(ffi.string(buf))
                    if i == #views and cb then
                        cb(ecode, buf)
                    end
                end, xml)
        end
    end
end

-- Function used to process the response to a get view list transaction request
function Configure:_handle_get_view_list_response(ecode, buf, all_errors)
    self.view_list_parsed = lxp.lom.parse(ffi.string(buf))
    self:_get_views()
end

-- Function used to retrieve the list of views from the device
function Configure:_get_view_list(ecode, data, all_errors)
    self:_send_transaction_request(C.TR_XML, C.TR_DUNNO_1,
        self._handle_get_view_list_response)
end

-- Handler called when we receive a response to the get channel list
-- transaction request
function Configure:_handle_get_channel_list_response(ecode, buf, all_errors)
    local xml = ffi.string(buf)
    local channel_list = lxp.lom.parse(xml)
    self.num_channels = 0
    for i = 2, #channel_list do
        local ref = channel_list[i][1]
        -- Don't include the "low speed" channels in the channel list
        if ref and ref:find('SyncSignal%[@address="1"%]') then
            self.num_channels = self.num_channels + 1
        elseif ref and ref:find('CanInChannel%[@address="1"%]') then
            self.num_channels = self.num_channels + 1
        end
    end
    self:_get_view_list()
end

-- Function used to retrieve a list of channel references
function Configure:get_reference_list()
    local ref_list = {}
    for _, c in ipairs(self.channel_info) do
        ref_list[#ref_list + 1] = c._reference
    end
    return ref_list
end

-- Function used to retrieve the channel list from the device
function Configure:_get_channel_list(ecode, data, all_errors)
    self:_send_transaction_request(C.TR_CHANNEL_REQUEST, C.TR_CHANNEL_LIST,
        self._handle_get_channel_list_response)
end


-- Function used to generate the next label to use in a transaction request or
-- an address resolution request
function Configure:get_label()
    self.label = self.label or 0
    self.label = bit.band(self.label + 1, 0xffff)
    return self.label
end

-- Function used to process address resolution reponse
function Configure:process_ar_response(buf, size)
    local comm_hdr = ffi.cast('struct communications_hdr *', buf)
    local dlen = ntohs(comm_hdr.payload_length) -
        ffi.sizeof('struct address_resolution_hdr')
    local data = buf + ffi.sizeof('struct communications_hdr') +
        ffi.sizeof('struct address_resolution_hdr')
    if dlen ~= ffi.sizeof('struct adr_response_payload') then
        self.logger:log(self.logger.ERR,
            "No data received in address response message")
        return
    end
    local arp = ffi.cast('struct adr_response_payload *', data)
    self.remote_type = ntohl(arp.type)
    self.remote_uuid = arp.uuid
    return true
end

-- Function used to process address resolution device type request
function Configure:process_ar_request_device_type(buf, size)
self.logger:log(self.logger.ERR, "process_ar_request_device_type")
    return true
end

-- Function used to process address resolution reponse pending
function Configure:process_ar_response_pending(buf, size)
self.logger:log(self.logger.ERR, "process_ar_response_pending")
    return true
end

Configure.process_ar_message = {}

Configure.process_ar_message[C.ADR_RESPONSE] =
    Configure.process_ar_response
Configure.process_ar_message[C.ADR_RESPONSE_PENDING] =
    Configure.process_ar_response_pending
Configure.process_ar_message[C.ADR_REQUEST_DEVICE_TYPE] =
    Configure.process_ar_request_device_type

-- Function used to process address resolution messages
function Configure:_process_address_resolution(buf, size)
    local ar_hdr = ffi.cast('struct address_resolution_hdr *',
        buf + ffi.sizeof('struct communications_hdr'))
    local process_message = self.process_ar_message[ntohs(ar_hdr.type)]
    if not process_message then
        self.logger:log(self.logger.ERR,
            "Unknown address request type in response")
        return
    end
    if not process_message(self, buf, size) then
        return
    end
    return size
end

local set_led_xml_template =[[
<?xml version="1.0"?>
<SrvParams>
<Control>LEDFLASH</Control>
<ControlParams>&lt;?xml version="1.0"
encoding="UTF-8"?&gt;&lt;ControlParams&gt;&lt;Par1&gt;0&lt;/Par1&gt;&lt;Par2&gt;#state#&lt;/Par2&gt;&lt;/ControlParams&gt;</ControlParams>
</SrvParams> 
]]  

-- Function used to set the state of the LED on the device
function Configure:set_led(state, cb)
    local xml = set_led_xml_template
    local state_table = {
        normal = '0',
        flashing_green = '1',
        flashing_orange = '2',
        flashing_red = '3',
    }
    xml = gsub_literal(xml, '#state#', state_table[state])
    -- I don't know why, but sometimes we don't get a response to this request
    -- even though the request is successful.
    self.devinfo.current_led_state = state
    self:_send_transaction_request(C.TR_SCHEMA, C.TR_XML,
        function(obj, ecode, data, all_errors)
            local result = generate_mx_results(
                'LED state change failed',
                xml, ecode, all_errors)
            if ecode ~= C.ERR_SUCCESS then
                self.logger:log(self.logger.ERR, result.msg)
            else
                self.devinfo.current_led_state = state
            end
            if cb then cb(result) end
        end, xml)
end

local reconnect_xml_template =[[
<?xml version="1.0"?>
<SrvParams>
<Control>LEDFLASH</Control>
<ControlParams>&lt;?xml version="1.0"
encoding="UTF-8"?&gt;&lt;ControlParams&gt;&lt;Par1&gt;#connector#&lt;/Par1&gt;&lt;/ControlParams&gt;</ControlParams>
</SrvParams>
]]

-- Function used to set the state of the LED on the device
function Configure:reconnect(connector, cb)
    local xml = reconnect_xml_template
    xml = gsub_literal(xml, '#connector#', tostring(connector))

    self:_send_transaction_request(C.TR_SCHEMA, C.TR_XML,
        function(obj, ecode, data, all_errors)
            local result = generate_mx_results(
                'Reconnect failed',
                xml, ecode, all_errors)
            if ecode ~= C.ERR_SUCCESS then
                self.logger:log(self.logger.ERR, result.msg)
            end
            if cb then cb(result) end
        end, xml)
end

local read_teds_binary_xml_template =
[[<?xml version="1.0" encoding="UTF-8"?>
<SrvParams><Channel>#connector#</Channel></SrvParams>]]

-- Function used to read the TEDS binary blob from a connector
function Configure:read_teds_binary(connector, cb)
    local xml = read_teds_binary_xml_template

    -- The connector numbers for this request is 0 (zero) based
    connector = tonumber(connector) - 1
    xml = gsub_literal(xml, '#connector#', tostring(connector))

    self:_send_transaction_request(C.TR_CHANNEL_DATA, C.TR_READ_TEDS_BINARY,
        function(obj, ecode, data, all_errors)
            local result = generate_mx_results(
                'Read TEDS binary blob failed',
                xml, ecode, all_errors)
            if ecode ~= C.ERR_SUCCESS then
                self.logger:log(self.logger.ERR, result.msg)
            end
--print(ecode, data, result)
            if cb then cb(result, data) end
        end, xml)
end

local write_teds_binary_xml_template =
[[<?xml version="1.0" encoding="UTF-8"?>
<SrvParams><Channel>#connector#</Channel><Tid>1</Tid><BinData>#bin_data#</BinData></SrvParams>]]

-- Function used to write bin data to a TEDS device on a connector
function Configure:write_teds_binary(connector, bin_data, cb)
    local xml = write_teds_binary_xml_template

    -- The connector numbers for this request is 0 (zero) based
    connector = tonumber(connector) - 1
    xml = gsub_literal(xml, '#connector#', tostring(connector))
    xml = gsub_literal(xml, '#bin_data#', bin_data)

    self:_send_transaction_request(C.TR_CHANNEL_DATA, C.TR_WRITE_TEDS_BINARY,
        function(obj, ecode, data, all_errors)
            local result = generate_mx_results(
                'Write TEDS binary blob failed',
                xml, ecode, all_errors)
            if ecode ~= C.ERR_SUCCESS then
                self.logger:log(self.logger.ERR, result.msg)
            else
                self:reconnect(connector, function(res)
                        self:set_channel_teds_enable(connector + 1, false,
                            function()
                                self:set_channel_teds_enable(connector + 1,
                                    true, function()
                                            if cb then cb(res, bin_data) end
                                    end)
                            end)
                    end)
            end
            if cb then cb(result, bin_data) end
        end, xml)
end

-- Function used to set the sample rate of a channel
function Configure:set_channel_sample_rate(channel_num, value, cb)
    local spec = {sample_rate = value}
    return self:setup_channel(channel_num, spec, cb)
end

-- Function used to enable/disable TEDS for a channel
function Configure:set_channel_teds_enable(channel_num, value, cb)
    -- NOTE:  the value for use_teds is 0 = Ignore, 2 = Auto, 3 = Required
    local spec = {
        _disable_filter_config = true,
        teds_sensor = {use_teds = (value and 2 or 0)}
    }
    return self:setup_channel(channel_num, spec, cb)
end

-- Function used to set the zero offset of a channel.  Note, this assume zero
-- offset stuff is enabled!
function Configure:set_channel_zero_offset(channel_num, value, cb)
    local template = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/ZeroOffset/Offset'
    template = gsub_literal(template, '#channel_num#', tostring(channel_num))

    local xml = string.format('<?xml version="1.0" encoding="UTF-8"?>\n<SrvParams><View>5</View><Path>%s</Path><XmlSection>%s</XmlSection></SrvParams>', template, tostring(value))

    self:_send_transaction_request(C.TR_DUNNO_0, C.TR_SET_CHANNEL_PARAMETER,
        function(obj, ecode, data, all_errors)
            local results = generate_mx_results(
                string.format('Unable to set zero offset for channel %d', chan_num),
                xml, ecode, all_errors)
            if cb then cb(results) end
        end, xml)

    return {result = 'success', msg = 'Channel configuration in progress'}
end
    
-- Function used to set the power LED to flashing green
function Configure:_flash_power_led(color)
    self.comm_not_working = nil
    self:set_led('flashing_green',
        function(result)
            if result.result == 'success' then
                self:_get_channel_list()
            end
        end)
end

-- Function used to set the power LED to normal mode
function Configure:_normal_power_led()
    self:set_led('normal')
end

-- Function used to setup the control connection and start the whole
-- initialization state machine.
function Configure:setup_connection()
    local size = ffi.sizeof('struct communications_hdr') +
        ffi.sizeof('struct address_resolution_hdr') +
        ffi.sizeof('struct communications_tail')

    local buf = self.buf

    local comm_hdr = ffi.cast('struct communications_hdr *', buf)
    comm_hdr.start = htonl(C.START_OF_MESSAGE)
    comm_hdr.len = ffi.sizeof('struct communications_hdr')
    comm_hdr.next_hdr_type = C.NEXT_HDR_TYPE_ADDRESS_RESOLUTION
    comm_hdr.payload_length = htons(ffi.sizeof('struct address_resolution_hdr'))
    comm_hdr.uuid = self.my_uuid

    local adr_hdr = ffi.cast('struct address_resolution_hdr *', comm_hdr + 1)
    adr_hdr.type = htons(C.ADR_REQUEST_INIT)
    adr_hdr.label = self:get_label()

    local tail = ffi.cast('struct communications_tail *', adr_hdr + 1)
    tail.tail = htonl(C.END_OF_MESSAGE)

    -- Set the callback to be called when the response is received
    self.callback = self.callback or {}
    self.callback[tonumber(adr_hdr.label)] = self._flash_power_led

    -- For whatever reason sometimes MX devices will announce that they are
    -- ready and have all of the neccessary protocols available but when we go
    -- to use the HBM Protocol to start using the device the MX device never
    -- sends a response back.  So, we timeout if we don't see a response within
    -- 3 seconds of sending the request
    self.comm_not_working = true

    local written = self.sbuf:write(buf, size)
    if written ~= size then
        self.logger:log(self.logger.ERR,
            "Unable to write address negotiation stuff")
        return
    end

    -- The response timeout timer mentioned above
    ev.Timer.new(function(l, t, e)
            if self.comm_not_working then
                -- If comm isn't working the MX device!
                log:log(log.NOTICE,
                    "HBM Protocol communications is not working " ..
                    "- rebooting device")
                local cmd = string.format(
                    "DROPBEAR_PASSWORD=hbmquantumx " ..
                    "/usr/bin/dbclient -y hbmuser@%s " ..
                    "'sync;sync;/sbin/reboot'", self.host)
                os.execute(cmd)
            end
        end, 3, 0):start(ev.Loop.default)
end

-- Function called when data arrives on the control connection
function Configure:_rx_message(available, buf)
    local offset = 0
    local d = buf
    while available > ffi.sizeof('struct communications_hdr') do
        local comm_hdr = ffi.cast('struct communications_hdr *', d)

        if comm_hdr.start ~= ntohl(C.START_OF_MESSAGE) then
            self.logger:log(self.logger.ERR,
                "Beginning of control channel buffer does not contain comm hdr")
            return offset
        end

        if comm_hdr.len ~= ffi.sizeof('struct communications_hdr') then
            self.logger:log(self.logger.ERR,
                "Unexpected header length in response")
            return offset
        end

        local plen = ntohs(comm_hdr.payload_length)

        local needed = ffi.sizeof('struct communications_hdr') + plen +
            ffi.sizeof('struct communications_tail')

        if available < needed then
            return offset
        end

        local comm_tail = ffi.cast('struct communications_tail *',
            d + needed - ffi.sizeof('struct communications_tail'))

        if comm_tail.tail ~= htonl(C.END_OF_MESSAGE) then
            self.logger:log(self.logger.ERR, string.format(
                "End of control channel buffer does not contain comm tail: %8X",
                comm_tail.tail))
            return offset
        end

        local nhdr_types = {
            [C.NEXT_HDR_TYPE_TRANSACTION] =
                'struct transaction_hdr *',
            [C.NEXT_HDR_TYPE_ADDRESS_RESOLUTION] =
                'struct address_resolution_hdr *',
        }
        local next_hdr = ffi.cast(nhdr_types[comm_hdr.next_hdr_type],
            d + ffi.sizeof('struct communications_hdr'))

        if comm_hdr.next_hdr_type == C.NEXT_HDR_TYPE_ADDRESS_RESOLUTION then
            if not self:_process_address_resolution(d, needed) then
                self.logger:log(self.logger.ERR,
                    "Unable to process address resolution message")
                return offset
            end
            local cb = self.callback[tonumber(next_hdr.label)]
            if cb then
                self.callback[tonumber(next_hdr.label)] = nil
                cb(self, ffi.cast('char *', next_hdr + 1))
            end
        elseif comm_hdr.next_hdr_type == C.NEXT_HDR_TYPE_TRANSACTION then
            if not self:_process_transaction(d, needed) then
                self.logger:log(self.logger.ERR,
                    "Unable to process transaction message")
                return offset
            end
        else
            self.logger:log(self.logger.ERR,
                string.format("Unknown next header type: %d",
                comm_hdr.next_hdr_type))
            return offset
        end
        available = available - needed
        offset = offset + needed
        d = d + needed
    end
    return offset
end

-- Function used to initialize a control connection to the device
function Configure:initialize(ip, logger, loop, cb)
    self.host = ip
    self.logger = logger
    self.loop = loop
    self.configuration_finished_cb = cb
    self.devinfo = {}

    -- Get the common parameter translators
    self.parameter_translator = {}
    for k, v in pairs(trans.common) do
        self.parameter_translator[k] = v
    end

    self._handle_get_view = self._handle_get_view_init

    self.buf = ffi.new('unsigned char[65536]')
    self.my_uuid = ffi.new('int64_t', math.random(0xffffffffffffffff))

    self.socket = net.socket{host = ip, port = 5001}
    local retries = 10
    while true do
        local ok, _ = pcall(self.socket.connect, self.socket)
        if ok then break end
        if retries == 0 then
            error()
        else
            retries = retries - 1
            C.sleep(2)
        end
    end

    local conn_stream_buf = net.StreamBuffer:new(self.socket.sock, (2^19))
    self.sbuf = conn_stream_buf
    local me = self

    function self.sbuf:read_cb(loop)
        local available, buf = conn_stream_buf:read()
        return me:_rx_message(available, buf)
    end

    function self.sbuf:stop_cb()
    end

    self.sbuf:start(loop)

    self:setup_connection()
end

function Configure:set_view_data(view_num, xpath, value, callback)

    local xml = xml_declaration ..
        '<SrvParams>' ..
            '<View>' .. view_num .. '</View>' ..
            '<Path>' .. xpath .. '</Path>' ..
            '<XmlSection>' .. value .. '</XmlSection>' ..
        '</SrvParams>'

    self:_send_transaction_request(C.TR_DUNNO_0, C.TR_CHANNEL_LIST,
        function(obj, ecode, data, all_errors)
            local results = generate_mx_results(
                'Unable to set MX configuration',
                xml, ecode, all_errors)
            if callback then callback(results) end
        end, xml)

end

function Configure:get_cal_cert(cb)

    local xml = xml_declaration ..
        '<SrvParams>' ..
            '<FilePath>/par/CalibrationCertificates.zip</FilePath>' ..
        '</SrvParams>'

    self:_send_transaction_request(6, 19,
        function(obj, ecode, data, all_errors)
            local results = generate_mx_results(
                'Unable to request certificate',
                xml, ecode, all_errors)
            if results.result == 'success' then
                self:_send_transaction_request(6, 21,
                    function(obj, ecode, data, all_errors)
                        local results = generate_mx_results(
                            'Unable to download certificate',
                            xml, ecode, all_errors)
                        if results.result == 'success' then
                            local pdf = ffi.string(data)
                            if cb then cb(true, '' .. data) end
                        else
                            if cb then cb(false, results.msg) end
                        end
                    end, '')
            else
                if cb then cb(false, results.msg) end
            end
        end, xml)

end

-- Function used to shutdown the control connection to the device
function Configure:close()
    self.sbuf:shutdown()
end

return Configure
