require 'strict'
local ffi = require 'ffi'

-- sie_utils.lua - Utility functions for the test engine to use in
--     setting up sie file XML.

local cjson = require 'cjson'
local log = require 'log'
local rational = require 'rational'
local sample_clock = require 'sample_clock'
local stempls = require 'test_engine/sie_templates'
local sie_schema = require 'test_engine/sie_schema'

-- "constants"
local NL = '\n'
local INDENT_LEV_1 = ' '

local Digital_Filters =
{
    "none",
    "bessel_iir",
    "butterworth_iir",
    "linear_phase_fir",
    "butterworth_fir"
}

-- Text buffer used to build blocks of channel tag XML.
-- NOTE:  This assumes m.write_chan_xml() is not used in
--        a reentrant fashion, so there is no competition
--        for use of this buffer.
local tag_buf = ffi.new('uint8_t[8192]')

-- Stores channel decoder ids.
local decoder_ids = {}

-- Stores info tables for defining base channels.
local ch_base_tab = {}

-- Stores info tables for defining base channels.
local histo_generics = {}

-- Builds a table with the information needed for a base channel.
-- NOTE that id fields are not set in the returned table,
--      specifically _sie_ch and _sie_decoder.
-- ALSO NOTE that table entries with keys WITHOUT a prepended '_'
--      will be written as channel tag elements!
-- Params:
--     'schema_info' contains information about the channel's schema.
-- Returns: table containing base channel info, followed by the
--     info's signature (the latter being a string built from the info)
local function base_chan_info(schema_info)
    local base_md = {}
    local sig = {}
    base_md._sie_name = "base channel"
    table.insert(sig, "_sie_name = " .. tostring(base_md._sie_name))
    base_md['somat:datamode_type'] = schema_info.somat_datamode_type
    table.insert(sig, "somat:datamode_type = " ..
        tostring(base_md['somat:datamode_type']))
    base_md['somat:data_bits'] = schema_info.bits
    table.insert(sig, "somat:data_bits = " ..
        tostring(base_md['somat:data_bits']))
    base_md['somat:data_format'] = schema_info.somat_data_format
    table.insert(sig, "somat:data_format = " ..
        tostring(base_md['somat:data_format']))
    base_md['core:description'] = ""
    table.insert(sig, "core:description = " ..
        tostring(base_md['core:description']))
    base_md['core:schema'] = schema_info.sie_core_schema
    table.insert(sig, "core:schema = " ..
        tostring(base_md['core:schema']))
    base_md['somat:version'] = "1.0"
    table.insert(sig, "somat:version = " ..
        tostring(base_md['somat:version']))

    return base_md, table.concat(sig, ';')
end

-- Finds the scaling mode from values in the stanza
local function get_scaling_mode(stanza)
    if type(stanza.calibration) == "table" then
        for _, el in ipairs(stanza.calibration) do
            if el.purpose == "sensor_scaling" then
                return tostring(el.mode)
            end
        end
    end
    return nil
end

-- Finds the appropriate scaling parameters and builds
-- them into the returned string.
local function get_scaling_parameters(mode, stanza)
    local step
    if mode == "internal" then
        return nil
    end
    local found = false
    local params = nil
    if type(stanza.calibration) == "table" then
        for _, v in ipairs(stanza.calibration) do
            if v.purpose == "sensor_scaling" then
                step = v
                found = true
                break
            end
        end
        if not found then
            return
        end
        if mode == "defined_slope_intercept" then
            params = string.format("m=%s,b=%s", tostring(step.slope or ""),
                tostring(step.intercept or ""))
        elseif mode == "defined_zero_span" then
            params = string.format("ps=%s,es=%s,ez=%s",
                tostring(step.physical_span or ""),
                tostring(step.electrical_span or ""),
                tostring(step.electrical_zero or ""))
        elseif mode == "defined_two_point" or
               mode == "experimental_two_point" then
            params = string.format("p1=%s,p2=%s,e1=%s,e2=%s",
                tostring(step.physical_1 or ""),
                tostring(step.physical_2 or ""),
                tostring(step.electrical_1 or ""),
                tostring(step.electrical_2 or ""))
        elseif mode == "strain_gage" then
            params = string.format("gf=%s,bf=%s",
                tostring(step.gage_factor or ""),
                tostring(step.bridge_factor or ""))
        elseif mode == "shunt" then
            params = string.format("res(kOhm)=%s,pol=%s,ps=%s,es=%s,ez=%s,%%dev=%s",
                tostring(step.resistor or ""), tostring(step.polarity or ""),
                tostring(step.physical_span or ""),
                tostring(step.electrical_span or ""),
                tostring(step.electrical_zero or ""),
                tostring(step.deviation_from_ideal or ""))
        else
            params = "Unknown mode"
        end
    end
    return params
end

-- Finds the appropriate "lead wire correction" parameters
-- and builds them into the returned string.
local function get_scaling_lead_wire_correction(mode, stanza)
    local step
    local found = false
    local params = nil
    if type(stanza.calibration) == "table" then
        for _, v in ipairs(stanza.calibration) do
            if v.purpose == "lead_wire_correction" then
                step = v
                found = true
                break
            end
        end
        if not found then
            return nil
        end
        if mode == "strain_gage" then
            params = string.format("gf=%s,bfc=%s",
                tostring(step.gage_factor or ""),
                tostring(step.bridge_factor_corrected or ""))
        elseif mode then
            params = string.format("m=%s,b=%s", tostring(step.slope or ""),
                tostring(step.intercept or ""))
        else
            params = "mode not found"
        end
    end
    return params
end

-- Returns string value for digital filter type
local function get_digital_filter_enum(filter_type)
    if filter_type ~= nil and type(filter_type) == "number" then
        return Digital_Filters[filter_type]
    else
        return nil
    end
end

-- Derives excitation carrier frequency from excitation mode
local function derive_excitation_freq(mode)
    if mode == "AC" then
        return "huh?"
    else
        return nil
    end
end

-- Derives excitation carrier wave form from excitation mode
local function derive_excitation_waveform(mode)
    if mode == "AC" then
        return "huh?"
    else
        return nil
    end
end

-- "constant" values for use by append_tag_line()
local TAG_PREFIX = ' <tag id="'
local TAG_PREFIX_LEN = #TAG_PREFIX
local TAG_CLOSE_BRACKET = '">'
local TAG_CLOSE_BRACKET_LEN = #TAG_CLOSE_BRACKET
local TAG_CLOSE_ELEMENT = '</tag>\n'
local TAG_CLOSE_ELEMENT_LEN = #TAG_CLOSE_ELEMENT

-- Used by build_channel_tags to create a single tag entry
-- in the temporary XML buffer.
-- Parameters:
--     'tag_buf' - the buffer in which to create the tag entry
--     'tag_len' - the current length of the XML string in 'tag_buf'
--     'key'     - the name to be used in the tag entry
--     'val'     - the value to be used in the tag entry
--     'empty_allowed'
--               - flag, indicates when true that a nil 'val'
--                 should be written as an empty string; when false 
--                 that no tag entry should be created
-- Returns the length of the XML string in 'tag_buf' after processing
local function xml_escape(str)
    return str:gsub('[&<>]', {
        ['&'] = '&amp;', ['<'] = '&lt;', ['>'] = '&gt;', })
end
local function append_tag_line(tag_buf, tag_len, key, val, empty_allowed)
    if val ~= nil then val = tostring(val) end
    if empty_allowed or (val ~= nil and #val > 0) then
        local key_str = tostring(key)
        if val == nil then val = "" end
        val = xml_escape(val)
        ffi.copy(tag_buf + tag_len, TAG_PREFIX, TAG_PREFIX_LEN)
        tag_len = tag_len + TAG_PREFIX_LEN
        ffi.copy(tag_buf + tag_len, key_str, #key_str)
        tag_len = tag_len + #key_str
        ffi.copy(tag_buf + tag_len, TAG_CLOSE_BRACKET, TAG_CLOSE_BRACKET_LEN)
        tag_len = tag_len + TAG_CLOSE_BRACKET_LEN
        ffi.copy(tag_buf + tag_len, val, #val)
        tag_len = tag_len + #val
        ffi.copy(tag_buf + tag_len, TAG_CLOSE_ELEMENT, TAG_CLOSE_ELEMENT_LEN)
        tag_len = tag_len + TAG_CLOSE_ELEMENT_LEN
    end
    return tag_len
end

-- Creates a buffer containing a channel's tag elements.
local function build_channel_tags(schema_info, stanza)
    assert(schema_info and stanza,
           "Internal error: bad parameter(s) to build_channel_tags")
    local tag_len = 0
    local summary = stanza.summary
    if type(summary) ~= 'table' then
        log:log(log.NOTICE, "MISSING stanza.summary")
        return tag_len, tag_buf
    end
    tag_len = append_tag_line(tag_buf, tag_len, 'core:uuid',
        schema_info['core:uuid'] or stanza.id, true)
    tag_len = append_tag_line(tag_buf, tag_len, 'data_type',
        stanza.data_type, true)
    tag_len = append_tag_line(tag_buf, tag_len, 'somat:data_format',
        schema_info.somat_data_format, true)
    tag_len = append_tag_line(tag_buf, tag_len, 'somat:message_format',
        stanza.message_format, false)
    tag_len = append_tag_line(tag_buf, tag_len, 'core:description',
        stanza.description, true)
    tag_len = append_tag_line(tag_buf, tag_len, 'somat:input_channel',
        stanza.somat_input_channel, true)
    tag_len = append_tag_line(tag_buf, tag_len, 'core:sample_rate',
        rational:new(stanza.sample_rate):tonumbers(), true)
    tag_len = append_tag_line(tag_buf, tag_len, 'somat:datamode_name',
        stanza.datamode_name or '__dm', true)
    tag_len = append_tag_line(tag_buf, tag_len, 'somat:datamode_type',
        schema_info.somat_datamode_type, true)
    tag_len = append_tag_line(tag_buf, tag_len, 'somat:connector',
        stanza.connector, true)
    tag_len = append_tag_line(tag_buf, tag_len, 'somat:data_bits',
        schema_info.bits, true)
    tag_len = append_tag_line(tag_buf, tag_len, 'core:schema',
        schema_info.sie_core_schema, true)
    tag_len = append_tag_line(tag_buf, tag_len,
        'somat:user_interface_sample_rate',
        stanza.sample_rate_legacy_uncorrected and
        stanza.sample_rate_legacy_uncorrected or nil, false)
    tag_len = append_tag_line(tag_buf, tag_len, 'somat:connection',
        stanza.connection and stanza.connection or nil, false)
    local mod_type = ""
    if stanza.connector then
        mod_type = stanza.connector:match("%.(%w+)%-")
    end
    tag_len = append_tag_line(tag_buf, tag_len, 'somat:module_type',
        mod_type, false)
    tag_len = append_tag_line(tag_buf, tag_len, 'somat:network_node',
        stanza.network_node and stanza.network_node or nil, false)
    tag_len = append_tag_line(tag_buf, tag_len, 'somat:measurand',
        stanza.measurand and stanza.measurand or nil, false)
    if summary.user_info then
        local custom_name = summary.user_info.custom_column_1 and
            summary.user_info.custom_column_1 or nil
        tag_len = append_tag_line(tag_buf, tag_len,
            'somat:user_defined_column_name_1', custom_name, false)
        tag_len = append_tag_line(tag_buf, tag_len,
            'somat:user_defined_column_value_1',
            custom_name and stanza.custom_column_1 and
            stanza.custom_column_1 or nil, custom_name ~= nil)
        custom_name = summary.user_info.custom_column_2 and
            summary.user_info.custom_column_2 or nil
        tag_len = append_tag_line(tag_buf, tag_len,
            'somat:user_defined_column_name_2', custom_name, false)
        tag_len = append_tag_line(tag_buf, tag_len,
            'somat:user_defined_column_value_2',
            custom_name and stanza.custom_column_2 and
            stanza.custom_column_2 or nil, custom_name ~= nil)
        custom_name = summary.user_info.custom_column_3 and
            summary.user_info.custom_column_3 or nil
        tag_len = append_tag_line(tag_buf, tag_len,
            'somat:user_defined_column_name_3', custom_name, false)
        tag_len = append_tag_line(tag_buf, tag_len,
            'somat:user_defined_column_value_3',
            custom_name and stanza.custom_column_3 and
            stanza.custom_column_3 or nil, custom_name ~= nil)
        custom_name = summary.user_info.custom_column_4 and
            summary.user_info.custom_column_4 or nil
        tag_len = append_tag_line(tag_buf, tag_len,
            'somat:user_defined_column_name_4', custom_name, false)
        tag_len = append_tag_line(tag_buf, tag_len,
            'somat:user_defined_column_value_4',
            custom_name and stanza.custom_column_4 and
            stanza.custom_column_4 or nil, custom_name ~= nil)
    end
    tag_len = append_tag_line(tag_buf, tag_len, 'somat:electrical_units',
        stanza.electrical_units and stanza.electrical_units or nil,
        false)
    tag_len = append_tag_line(tag_buf, tag_len, 'somat:output_units',
        stanza.output_units and stanza.output_units or nil, false)
    local scaling_mode = get_scaling_mode(stanza)
    tag_len = append_tag_line(tag_buf, tag_len, 'somat:scaling_mode',
        scaling_mode, false)
    tag_len = append_tag_line(tag_buf, tag_len, 'somat:scaling_parameters',
        get_scaling_parameters(scaling_mode, stanza), false)
    tag_len = append_tag_line(tag_buf, tag_len,
        'somat:scaling_lead_wire_correction',
        get_scaling_lead_wire_correction(scaling_mode, stanza), false)
    tag_len = append_tag_line(tag_buf, tag_len, 'somat:zero_offset_target',
        stanza.prerun_balance and stanza.prerun_balance.target_value and
        stanza.prerun_balance.target_value or nil, false)
    tag_len = append_tag_line(tag_buf, tag_len, 'somat:zero_offset_value',
        stanza.prerun_balance and stanza.prerun_balance.offset and
        stanza.prerun_balance.offset or nil, false)
    tag_len = append_tag_line(tag_buf, tag_len, 'somat:physical_range_min',
        stanza.range_min and stanza.range_min or nil, false)
    tag_len = append_tag_line(tag_buf, tag_len, 'somat:physical_range_max',
        stanza.range_max and stanza.range_max or nil, false)
    local digital_filter_type_nbr = stanza.digital_filter_type or nil
    local digital_filter_type_str = digital_filter_type_nbr and
        stanza.digital_filter_type or nil
    tag_len = append_tag_line(tag_buf, tag_len, 'somat:digital_filter_type',
        digital_filter_type_str, false)
    tag_len = append_tag_line(tag_buf, tag_len,
        'somat:digital_filter_attenuation_frequency',
        stanza.digital_filter_attenuation_frequency and
        stanza.digital_filter_attenuation_frequency or nil, false)
    tag_len = append_tag_line(tag_buf, tag_len, 'hbm_enum:digital_filter_type',
        get_digital_filter_enum(digital_filter_type_nbr), false)
    local excitation_mode =
        stanza.excitation_mode and stanza.excitation_mode or nil
    tag_len = append_tag_line(tag_buf, tag_len, 'somat:excitation_mode',
        excitation_mode, false)
    tag_len = append_tag_line(tag_buf, tag_len,
        'somat:ac_excitation_carrier_frequency',
        derive_excitation_freq(excitation_mode), false)
    tag_len = append_tag_line(tag_buf, tag_len,
        'somat:ac_excitation_carrier_waveform',
        derive_excitation_waveform(excitation_mode), false)
    tag_len = append_tag_line(tag_buf, tag_len, 'somat:excitation_range',
        stanza.excitation_range and stanza.excitation_range or nil,
        false)
    tag_len = append_tag_line(tag_buf, tag_len, 'somat:transducer_power_voltage',
        stanza.transducer_power and stanza.transducer_power or nil,
        false)
    local sensor_type = stanza.sensor_input and stanza.sensor_input.mode or nil
    tag_len = append_tag_line(tag_buf, tag_len, 'somat:sensor_type',
        sensor_type, false)
    tag_len = append_tag_line(tag_buf, tag_len, 'somat:electrical_range_min',
        stanza.sensor_input and stanza.sensor_input.min or nil, false)
    tag_len = append_tag_line(tag_buf, tag_len, 'somat:electrical_range_max',
        stanza.sensor_input and stanza.sensor_input.max or nil, false)
    if sensor_type == "bridge" then
        tag_len = append_tag_line(tag_buf, tag_len, 'somat:bridge_type',
            stanza.sensor_input["type"] or nil, false)
        tag_len = append_tag_line(tag_buf, tag_len, 'somat:bridge_wires',
            stanza.sensor_input.wires or nil, false)
        tag_len = append_tag_line(tag_buf, tag_len, 'somat:bridge_resistance',
            stanza.bridge_resistance or nil, false)
        tag_len = append_tag_line(tag_buf, tag_len,
            'somat:bridge_lead_wire_resistance',
                    stanza.bridge_lead_wire_resistance or nil, false)
    elseif sensor_type == "inductive_bridge" then
        tag_len = append_tag_line(tag_buf, tag_len, 'somat:bridge_type',
            stanza.sensor_input["type"] or nil, false)
        tag_len = append_tag_line(tag_buf, tag_len, 'somat:bridge_wires',
            stanza.sensor_input.wires or nil, false)
    elseif sensor_type == "strain_gage" or sensor_type == "potentiometer" then
        tag_len = append_tag_line(tag_buf, tag_len, 'somat:bridge_wires',
            stanza.sensor_input.wires or nil, false)
    end
    tag_len = append_tag_line(tag_buf, tag_len, 'somat:thermocouple_type',
        (sensor_type == "thermocouple") and
                stanza.sensor_input["type"] or nil, false)
    if stanza.sensor_input then
        tag_len = append_tag_line(tag_buf, tag_len, 'somat:encoder_quadrature',
            stanza.sensor_input.quadrature, false)
        tag_len = append_tag_line(tag_buf, tag_len, 'somat:encoder_track_sign',
            stanza.sensor_input.track_sign, false)
        tag_len = append_tag_line(tag_buf, tag_len, 'somat:encoder_index_reset',
            stanza.sensor_input.encoder_index_reset, false)
        tag_len = append_tag_line(tag_buf, tag_len, 'somat:encoder_index_divisor',
            stanza.sensor_input.encoder_index_divisor, false)
        tag_len = append_tag_line(tag_buf, tag_len, 'somat:encoder_pulses_per_round',
            stanza.sensor_input.pulses_per_round, false)
    end
    tag_len = append_tag_line(tag_buf, tag_len,
        'somat:stale_data_expiration_time', stanza.stale_data_expiration_time,
        false)
    tag_len = append_tag_line(tag_buf, tag_len,
        'somat:invalid_data_output_value', stanza.invalid_data_output_value,
        false)
    -- CAN related
    if stanza.db_bit_start then
        tag_len = append_tag_line(tag_buf, tag_len, 'somat:can_bit_rate',
            stanza.bit_rate, false)
        tag_len = append_tag_line(tag_buf, tag_len, 'somat:can_msg_id',
            stanza.db_msg_id, false)
        tag_len = append_tag_line(tag_buf, tag_len, 'somat:can_bit_length',
            stanza.db_bit_length, false)
        tag_len = append_tag_line(tag_buf, tag_len, 'somat:can_bit_start',
            stanza.db_bit_start, false)
        tag_len = append_tag_line(tag_buf, tag_len,
            'somat:can_data_format_string', stanza.db_data_format, false)
        tag_len = append_tag_line(tag_buf, tag_len, 'somat:can_offset',
            stanza.db_offset, false)
        tag_len = append_tag_line(tag_buf, tag_len, 'somat:can_scale',
            stanza.db_scale, false)
        tag_len = append_tag_line(tag_buf, tag_len, 'somat:can_max',
            stanza.db_max, false)
        tag_len = append_tag_line(tag_buf, tag_len, 'somat:can_min',
            stanza.db_min, false)
        tag_len = append_tag_line(tag_buf, tag_len, 'somat:can_units',
            stanza.db_units, false)
        if mod_type == 'cx23r' or mod_type == 'edaqxr' then
            tag_len = append_tag_line(tag_buf, tag_len, 'somat:can_request_rate',
                stanza.request_rate, false)
            tag_len = append_tag_line(tag_buf, tag_len, 'somat:can_request_msg',
                stanza.request_msg, false)
        end
    end
    if mod_type == 'cx23r' or mod_type == 'edaqxr' then
        tag_len = append_tag_line(tag_buf, tag_len,
            'somat:digital_threshold_high',
            stanza.threshold and stanza.threshold.max or nil, false)
        tag_len = append_tag_line(tag_buf, tag_len,
            'somat:digital_threshold_low',
            stanza.threshold and stanza.threshold.min or nil, false)
    end
    if mod_type == 'cx23r' or mod_type == 'mx460' then
        tag_len = append_tag_line(tag_buf, tag_len,
            'somat:internal_termination', stanza.internal_termination, false)
    elseif mod_type:match('^mx840') then
        if stanza.sensor_input and
           stanza.sensor_input.mode == 'ssi_encoder_counter' then
            tag_len = append_tag_line(tag_buf, tag_len,
                'somat:ssi_coding', stanza.sensor_input.ssi_coding, false)
            tag_len = append_tag_line(tag_buf, tag_len,
                'somat:ssi_clock', stanza.sensor_input.ssi_clock, false)
            tag_len = append_tag_line(tag_buf, tag_len,
                'somat:ssi_bits', stanza.sensor_input.ssi_bits, false)
        end
    end
    if mod_type == 'edaqxr' then
        tag_len = append_tag_line(tag_buf, tag_len,
            'somat:smart_module_type',
            stanza.smart_module and stanza.smart_module['type'] or nil, false)
        tag_len = append_tag_line(tag_buf, tag_len,
            'somat:smart_module_sn',
            stanza.smart_module and stanza.smart_module.serial or nil, false)
    end
    if stanza.teds_sensor and stanza.teds_sensor.bin_data then
        tag_len = append_tag_line(tag_buf, tag_len,
            'somat:teds_sensor_id', stanza.teds_sensor.id, false)
        tag_len = append_tag_line(tag_buf, tag_len,
            'somat:teds_calibration_date',
            stanza.teds_sensor.calibration_date, false)
        tag_len = append_tag_line(tag_buf, tag_len,
            'somat:teds_calibration_period',
            stanza.teds_sensor.calibration_period, false)
    end
    if stanza.database_sensor then
        tag_len = append_tag_line(tag_buf, tag_len,
            'somat:sensor_id', stanza.database_sensor.id, false)
        tag_len = append_tag_line(tag_buf, tag_len,
            'somat:sensor_calibration_date',
            stanza.database_sensor.calibration_date, false)
        tag_len = append_tag_line(tag_buf, tag_len,
            'somat:sensor_calibration_period',
            stanza.database_sensor.calibration_period, false)
    end

    assert(tag_len < 8192, "Internal error: channel tag temp buffer overflow")
    return tag_len, tag_buf
end

-- The information needed to write the SIE XML for a specific
-- test channel comprises a subset of the values from the channel's
-- setup stanza.  This function provides that subset.
-- NOTE that id fields are not set in the returned table,
--      specifically _sie_ch, _sie_group, and _sie_decoder.
--      However, _sie_group is set to 'true' if a group id
--      is needed, or to 'false' if a group id is not needed.
-- ALSO NOTE that channel tag elements are combined into one buffer
--      which can be written with one call to sie_writer:xml(),
--      for efficiency.
-- Params:
--     'stanza' contains setup info for this channel
-- Returns: table containing info for this channel
local function test_chan_info(schema_info, stanza)
    local ch_md = {}
    ch_md._sie_test = stanza._sie_test or 0
    ch_md._sie_name = stanza.sie_name
    if stanza.private_chan then
        ch_md._sie_group = false
    else
        ch_md._sie_group = true
    end
    ch_md._all_ch_tags_len, ch_md._all_ch_tags =
        build_channel_tags(schema_info, stanza)

    local srate = rational:new(stanza.sample_rate):tonumbers()
    if type(srate) == 'number' and srate > 0 then
        ch_md._scale = 1.0 / srate
        ch_md._offset = 0
    else
        ch_md._scale = 1.0 / sample_clock.master_rate
        ch_md._offset = 0
    end

    return ch_md
end

-- Puts base channel's id in local table, using base channel's
-- signature as the key.
-- Params:
--     'sig' signature of the base channel
--     'id' base channel's id
local function ch_base_tab_put(sig, id)
    ch_base_tab[sig] = id
end

-- Finds the id of the base channel which has the specified signature.
-- Params:  'sig' is the signature of the desired base channel
-- Returns: id of the base channel with the specified signature,
--     otherwise nil
local function ch_base_tab_get(sig)
    return ch_base_tab[sig]
end

-- Module table.
local m = {}

--  Values for sie buf objects' dm_type field.
--  These indicate sie bufs which are used for and need
--  special processing for a specific data mode.
--    For synchronous datamodes which are distinguished
--    only by having discontinuous timestamps:
m.DISCON_DATAMODE = 0x40
--    For a Block History Datamode:
m.BH_DATAMODE = 0x42

-- Converts the input into a rational number.
-- (This function was copied from sie_writer.lua)
function m.get_rational(x)
    local slash = (type(x) == 'string') and x:find('/')
    if slash then
        local num = tonumber(x:sub(1, slash - 1))
        local denom = tonumber(x:sub(slash + 1))
        return num / denom, num, denom
    end
    local num = tonumber(x)
    return num, num, 1
end

-- Adds some useful values to the given stanza
-- Parameters:
--     'stanza' is the (decoded) setup file stanza to modify
--     'datamode_name' is the name of the datamode associated
--         with the channel described by 'stanza' - may be nil
--         if there is no associated datamode
function m.supplement_stanza_info(stanza, datamode_name)
    stanza.sie_name = datamode_name .. '@' .. stanza.name
    stanza.datamode_name = datamode_name
    stanza.somat_input_channel = stanza.name
end

-- Writes XML to define a SIE decoder's loop.
-- Parameters:
--     'sw' is the sie_writer object to use
--     'k'  is the key value for the set of loop attributes,
--          which indicates what type of loop is specified
--     'v'  is the the set of loop attributes to be used
--     'indent'  is the level of indentation of the XML
function m.write_decoder_loop(sw, k, v, indent)
    -- write loop start
    if k:find('^bare') then
        sw:xml(INDENT_LEV_1:rep(indent) .. stempls.dec_bare_loop .. NL, v.head)
    elseif k:find('^short') then
        sw:xml(INDENT_LEV_1:rep(indent) .. stempls.dec_short_loop .. NL, v.head)
    elseif k:find('^long') then
        sw:xml(INDENT_LEV_1:rep(indent) .. stempls.dec_long_loop .. NL, v.head)
    end
    if v.inner then
        -- write inner loop
        m.write_decoder_loop(sw, k, v.inner, indent+1)
    end
    -- write interior reads/sets/sample
    for _, r in ipairs(v.reads or {}) do
        if r.bits then
            sw:xml(INDENT_LEV_1:rep(indent+1) .. stempls.dec_read .. NL, r)
        else
            sw:xml(INDENT_LEV_1:rep(indent+1) ..
                   stempls.dec_read_octets .. NL, r)
        end
        if r.post_seek then
            sw:xml(INDENT_LEV_1:rep(indent+1) .. stempls.dec_seek .. NL,
                   r.post_seek)
        end
    end
    if v.sample then
        sw:xml(INDENT_LEV_1:rep(indent+1) .. stempls.dec_sample .. NL)
    end
    for _, r in ipairs(v.sets or {}) do
        sw:xml(INDENT_LEV_1:rep(indent+1) .. stempls.dec_set .. NL, r)
    end
    -- write loop end
    sw:xml(INDENT_LEV_1:rep(indent) .. stempls.dec_loop_end .. NL)
end

-- Writes XML to define a SIE decoder.
-- Parameters:
--     'sw' is the sie_writer object to use
--     'schema_info' contains information about the somat
--         schema with which this decoder will be associated,
--         and which is needed to write the XML
-- Returns:  id for the newly written decoder
function m.write_decoder(sw, schema_info)
    local decoder_id = sw:next_id('decoder')
    sw:xml(stempls.dec_start .. NL, {_sie_decoder = decoder_id})
    for _, v in ipairs(schema_info.nonrep_reads or {}) do
        sw:xml(INDENT_LEV_1 .. stempls.dec_read .. NL, v)
    end
    for _, v in ipairs(schema_info.nonrep_sets or {}) do
        sw:xml(INDENT_LEV_1 .. stempls.dec_set .. NL, v)
    end
    for k, v in pairs(schema_info.loops or {}) do
        m.write_decoder_loop(sw, k, v, 1)
    end
    sw:xml(stempls.dec_end .. NL)
    return decoder_id
end

-- Writes the info in the given table to the SIE file as 
-- the XML for a channel.
-- Parameters:
--     'sw' is the sie_writer object to use
--     'md' contains the remaining info to be written
--     'dims' contains info about the channel's "dim" elements
--         (NOTE that this may be nil, for base channels)
function m.write_chan_xml(sw, md, dims)
    if md._priv_type then
        -- write a base channel
        sw:xml(stempls.base_ch .. NL, md)
    elseif md._sie_group then
        -- write a specific test channel
        sw:xml(stempls.test_ch .. NL, md)
    else
        -- write a specific private test channel
        sw:xml(stempls.priv_test_ch .. NL, md)
    end
    if md._all_ch_tags_len > 0 then
        -- write <tag> elements
        sw:raw_xml(ffi.string(md._all_ch_tags,
                   md._all_ch_tags_len),
                   md._all_ch_tags_len)
    end

    if dims then
        -- write <dim> elements
        for i, dim in ipairs(dims) do
            sw:xml(INDENT_LEV_1 .. stempls.dim_hdr .. NL, dim)
            -- NOTE:  Only keys WITHOUT prepended '_' will be written
            --        as dim tag elements!
            for k, v in pairs(dim) do
                if k:sub(1, 1) ~= '_' then
                    sw:xml(INDENT_LEV_1:rep(2) .. stempls.ch_tag .. NL,
                        {k=k, v=v})
                end
            end
            sw:xml(INDENT_LEV_1:rep(2) .. stempls.dec_assign .. NL, dim, md)
            if dim._use_scale then
                if dim._use_index then
                    sw:xml(INDENT_LEV_1:rep(2) .. stempls.dec_xindex .. NL,
                        {_index_ch = dim._index_ch,
                         _index_dim = dim._index_dim})
                else
                    sw:xml(INDENT_LEV_1:rep(2) .. stempls.dec_scale .. NL,
                        {_scale = md._scale,
                         _offset = md._offset})
                end
            end
            -- finish writing <dim> element
            sw:xml(INDENT_LEV_1 .. stempls.dim_end .. NL)
        end
    end
    -- finish writing channel element
    sw:xml(stempls.ch_end .. NL)
end

-- Writes the SIE XML for a channel's SIE base channel.
-- Parameters:
--     'sw' is the sie_writer object to use
--     'md' contains information about the base channel
--         NOTE that this table's id fields will be modified
--     'decoder_type' indicates the signature of the decoder
--         to be used
function m.write_channel_base(sw, md, decoder_type)
    md._decoder_type = decoder_type
    md._priv_type = "base"
    m.write_chan_xml(sw, md)
end

-- Writes the SIE XML for a specific test channel.
-- Parameters:
--     'sw' is the sie_writer object to use
--     'md' contains information about the channel
--     'schema_info' contains information about the
--         channel's schema.
function m.write_channel(sw, md, schema_info)
    m.write_chan_xml(sw, md, schema_info.dims)
end

-- Writes XML to the SIE file for a channel.  This includes
-- writing the decoder and/or base channel XML if needed.
-- Parameters:
--     'sw' is the sie_writer object to use
--     'schema_info' contains info about a channel's SIE schema
--     'stanza' contains setup info for a channel
-- Returns:  table containing this specific channel's info
function m.write_sie_channel(sw, schema_info, stanza)
    assert(sw, "Missing sw (sie_writer)")
    assert(schema_info, "Missing SIE schema info")
    assert(stanza, "Missing channel setup info")

    -- get from the schema info (most of) the values needed to
    -- write the XML for a base channel
    local base_md
    local base_sig
    -- information for base channel
    base_md, base_sig = base_chan_info(schema_info)
    -- information for this specific channel
    local chan_md = test_chan_info(schema_info, stanza)
    -- need decoder id
    local decoder_id = decoder_ids[schema_info.decoder_type]
    if not decoder_id then
        -- decoder doesn't already exist, so write XML for it
        decoder_id = m.write_decoder(sw, schema_info)
        decoder_ids[schema_info.decoder_type] = decoder_id
    end
--~     -- get base channel id
--~     local base_id = ch_base_tab_get(base_sig)
--~     if not base_id then
--~         base_id = sw:next_id('ch')
--~         -- base channel doesn't already exist, so write XML for it
--~         ch_base_tab_put(base_sig, base_id)
--~         -- set ids for the base channel
--~         base_md._sie_decoder = decoder_id
--~         base_md._sie_ch = base_id
--~         m.write_channel_base(sw, base_md, schema_info.decoder_type)
--~     end
    chan_md._sie_decoder = decoder_id
--~     chan_md._base_id = base_id
    chan_md._sie_ch = sw:next_id('ch')
    if chan_md._sie_group then
        chan_md._sie_group = sw:next_id('group')
    end
    m.write_channel(sw, chan_md, schema_info)
    return chan_md
end

-- Checks for the existence of the min/max lookup table base channel
-- needed for a histogram.  Writes it if it doesn't exist (includes
-- writing decoder XML).
-- Parameters:
--     'sw' is the sie_writer object to use
--     'bounds_bits' size in bits of bounds min/max datatype
-- Returns:  table containing this specific channel's info
function m.ensure_sie_bounds_channel(sw, bounds_bits)
    local bounds_type = '_histogram_bin_bounds' .. bounds_bits
    local ch_id = histo_generics[bounds_type]
    if not ch_id then
        -- write decoder
        local decoder_info = sie_schema.get_hist_decoder_schema(bounds_bits)
        local decoder_id = m.write_decoder(sw, decoder_info)

        -- write channel (start XML on new line)
        ch_id = sw:next_id('ch')
        sw:xml(NL .. stempls.generic_bin_ch .. NL, {_sie_ch = ch_id})
        sw:xml(INDENT_LEV_1 .. stempls.ch_tag .. NL, {k = "core:description",
            v = "Generic histogram bin min/max values lookup table"})
        sw:xml(INDENT_LEV_1 .. stempls.dim_hdr .. NL, {_dim_index = "0"})
        sw:xml(INDENT_LEV_1:rep(2) .. stempls.dec_assign .. NL,
            {_sie_decoder = decoder_id, _decoder_v = "0"})
        sw:xml(INDENT_LEV_1 .. stempls.dim_end .. NL)

        sw:xml(INDENT_LEV_1 .. stempls.dim_hdr .. NL, {_dim_index = "1"})
        sw:xml(INDENT_LEV_1:rep(2) .. stempls.dec_assign .. NL,
            {_sie_decoder = decoder_id, _decoder_v = "1"})
        sw:xml(INDENT_LEV_1 .. stempls.dim_end .. NL)

        sw:xml(INDENT_LEV_1 .. stempls.dim_hdr .. NL, {_dim_index = "2"})
        sw:xml(INDENT_LEV_1:rep(2) .. stempls.dec_assign .. NL,
            {_sie_decoder = decoder_id, _decoder_v = "2"})
        sw:xml(INDENT_LEV_1 .. stempls.dim_end .. NL)

        sw:xml(stempls.ch_end .. NL)
        sw:xml()
        histo_generics[bounds_type] = ch_id
    end
    return ch_id
end

-- Writes the XML to describe a channel which is used to decode
-- one of the bin bounds dimensions for a histogram.
-- Parameters:
--     'sw' is the sie_writer object to use
--     'md' data needed for the XML template
-- Returns:  a new channel id created for this bin channel
function m.write_bin_channel(sw, md)
    md._sie_ch = sw:next_id('ch')
    sw:xml(stempls.priv_bin_ch .. NL, md)
    sw:xml()
    return md._sie_ch
end

-- Writes the XML for a Rainflow datamode's unclosed cycles data.
-- This occurs at the end of a test.  The cycles data is converted
-- to ASCII floating point format before writing.
-- Parameters:
--     'sw' is the sie_writer object to use
--     'md' data needed for the XML template
--     'cycles' array containing unclosed cycles data, in raw 64 bit format
--     'len' number of elements in 'cycles' array
-- Returns:  a new channel id created for this bin channel
function m.write_unclosed_cycles(sw, md, cycles, len)
    local asciiCycles = {}

    -- convert cycles data
    for i = 1, len do
        asciiCycles[#asciiCycles + 1] = string.format("%f", cycles[i])
    end
    -- write XML with converted data
    md._unclosed_cycles = table.concat(asciiCycles, " ")
    sw:xml(NL .. stempls.unclosed_cycles .. NL, md)
end

return m

